import numpy as np

from beluga.utils import timeout, keyboard
from beluga.bvpsol.algorithms.BaseAlgorithm import BaseAlgorithm
from beluga.ivpsol import Propagator
import copy
import numba

import logging
import itertools as it


class Shooting(BaseAlgorithm):
    """
    Shooting algorithm for solving boundary value problems.

    Given a system of ordinary differential equations :eq:`ordinarydifferentialequation`, define the sensitivities as

    .. math::
        A(t) = \\left[\\frac{\\partial \\mathbf{f}}{\\partial \\mathbf{x}}, \\frac{\\partial \\mathbf{f}}{\\partial \\mathbf{p}}\\right]

    Then, the state-transition matrix is defined as the following set of first-order differential equations

    .. math::
        \\begin{aligned}
            \\Delta_0 &= \\left[Id_M, \\mathbf{0}\\right] \\\\
            \\dot{\\Delta} &= A(t)\\Delta
        \\end{aligned}

    Sensitivities of the boundary conditions are

    .. math::
        \\begin{aligned}
            M &= \\frac{\\partial \mathbf{\\Phi}}{\\partial \\mathbf{x}_0} \\\\
            P &= \\frac{\\partial \mathbf{\\Phi}}{\\partial \\mathbf{p}} \\\\
            Q_0 &= \\frac{\\partial \mathbf{\\Phi}}{\\partial \\mathbf{q}_0} \\\\
            Q_f &= \\frac{\\partial \mathbf{\\Phi}}{\\partial \\mathbf{q}_f}
        \\end{aligned}

    The Jacobian matrix is then the concatenation of these sensitivities

    .. math::
        J = \\left[M, P, Q_0+Q_f \\right]
    """

    def __new__(cls, *args, **kwargs):
        """
        Creates a new Shooting object.

        :param args: Unused
        :param kwargs: Additional parameters accepted by the solver.
        :return: Shooting object.

        +------------------------+-----------------+-----------------+
        | Valid kwargs           | Default Value   | Valid Values    |
        +========================+=================+=================+
        | cached                 | True            | Bool            |
        +------------------------+-----------------+-----------------+
        | derivative_method      | 'fd'            | {'fd','csd'}    |
        +------------------------+-----------------+-----------------+
        | ivp_args               | {}              | see `ivpsol`    |
        +------------------------+-----------------+-----------------+
        | tolerance              | 1e-4            | > 0             |
        +------------------------+-----------------+-----------------+
        | max_error              | 100             | > 0             |
        +------------------------+-----------------+-----------------+
        | max_iterations         | 100             | > 0             |
        +------------------------+-----------------+-----------------+
        | use_numba              | False           | Bool            |
        +------------------------+-----------------+-----------------+
        | verbose                | False           | Bool            |
        +------------------------+-----------------+-----------------+
        """

        obj = super(Shooting, cls).__new__(cls, *args, **kwargs)

        cached = kwargs.get('cached', True)
        derivative_method = kwargs.get('derivative_method', 'fd')
        ivp_args = kwargs.get('ivp_args', dict())
        tolerance = kwargs.get('tolerance', 1e-4)
        max_error = kwargs.get('max_error', 100)
        max_iterations = kwargs.get('max_iterations', 100)
        use_numba = kwargs.get('use_numba', False)
        verbose = kwargs.get('verbose', False)

        obj.cached = cached
        obj.derivative_method = derivative_method
        obj.ivp_args = ivp_args
        obj.tolerance = tolerance
        obj.max_error = max_error
        obj.max_iterations = max_iterations
        obj.use_numba = use_numba
        obj.verbose = verbose
        return obj

    def __init__(self, *args, **kwargs):
        self.stm_ode_func = None
        self.saved_code = True

        if self.derivative_method not in ['fd']:
            raise ValueError("Invalid derivative method specified. Valid options are 'csd' and 'fd'.")

    def preprocess(self, problem_data, use_numba=False):
        obj = super().preprocess(problem_data, use_numba=use_numba)
        self.stm_ode_func = self.make_stmode(obj[0].deriv_func, problem_data['nOdes'])
        return obj

    def _bc_jac_multi(self, nBCs, phi_full_list, y_list, parameters, aux, bc_func, StepSize=1e-6):
        p  = np.array(parameters)
        nParams = p.size
        h = StepSize
        ya = np.array([traj.T[0] for traj in y_list]).T
        yb = np.array([traj.T[-1] for traj in y_list]).T
        nOdes = ya.shape[0]
        num_arcs = len(phi_full_list)
        fx = bc_func(ya,yb,p,aux)
        nBCs = len(fx)

        M = np.zeros((nBCs, nOdes))
        P = np.zeros((nBCs, p.size))
        J = np.zeros((nBCs, (nOdes)*num_arcs+p.size))
        dx = np.zeros((nOdes+nParams, num_arcs))

        for arc_idx, phi in zip(it.count(), phi_full_list):
            # Evaluate for all arcs
            for i in range(nOdes):
                dx[i, arc_idx] = dx[i, arc_idx] + h
                dy = np.dot(phi[-1], dx)
                f = bc_func(ya + dx[:nOdes], yb + dy, p, aux)
                M[:,i] = (f-fx)/h
                dx[i, arc_idx] = dx[i, arc_idx] - h

            J_i = M
            J_slice = slice(nOdes*arc_idx, nOdes*(arc_idx+1))
            J[:,J_slice] = J_i

        for i in range(p.size):
            p[i] = p[i] + h
            j = i + nOdes
            dx[j, arc_idx] = dx[j, arc_idx] + h
            dy = np.dot(phi[-1], dx)
            f = bc_func(ya, yb + dy, p, aux)
            P[:,i] = (f-fx)/h
            dx[j, arc_idx] = dx[j, arc_idx] - h
            p[i] = p[i] - h

        # J[:,nOdes*num_arcs:] = P
        J = np.hstack((M,P))
        return J

    @staticmethod
    def make_stmode(odefn, nOdes, StepSize=1e-6):
        Xh = np.eye(nOdes)*StepSize

        # @numba.jit(looplift=True, nopython=True)
        def _stmode_fd(t, _X, p, const, arc_idx):
            """ Finite difference version of state transition matrix """
            nParams = p.size
            F = np.empty((nOdes, nOdes+nParams))
            phi = _X[nOdes:].reshape((nOdes, nOdes+nParams))
            X = _X[0:nOdes]  # Just states

            # Compute Jacobian matrix, F using finite difference
            fx = np.squeeze([odefn(t, X, p, const, arc_idx)])

            for i in numba.prange(nOdes):
                fxh = odefn(t, X + Xh[i, :], p, const, arc_idx)
                F[:, i] = (fxh-fx)/StepSize

            for i in numba.prange(nParams):
                p[i] += StepSize
                fxh = odefn(t, X, p, const, arc_idx)
                F[:, i+nOdes] = (fxh - fx).real / StepSize
                p[i] -= StepSize

            phiDot = np.dot(np.vstack((F, np.zeros((nParams, nParams + nOdes)))), np.vstack((phi, np.hstack((np.zeros((nParams, nOdes)), np.eye(nParams))))))[:nOdes, :]
            return np.concatenate((fx, np.reshape(phiDot, (nOdes * (nOdes + nParams)))))

        def wrapper(t, _X, p, const, arc_idx):  # needed for scipy
            return _stmode_fd(t, _X, p, const, arc_idx)
        return wrapper

    def solve(self, deriv_func, quad_func, bc_func, solinit):
        """
        Solve a two-point boundary value problem using the shooting method.

        :param deriv_func: The ODE function.
        :param quad_func: The quad func.
        :param bc_func: The boundary conditions function.
        :param solinit: An initial guess for a solution to the BVP.
        :return: A solution to the BVP.
        """

        # Instantly make a copy of sol and format inputs
        sol = copy.deepcopy(solinit)
        sol.t = np.array(sol.t, dtype=np.float64)
        sol.y = np.array(sol.y, dtype=np.float64)
        sol.parameters = np.array(sol.parameters, dtype=np.float64)

        # Extract some info from the guess structure
        y0g = sol.y[:, 0]
        nOdes = y0g.shape[0]
        paramGuess = sol.parameters

        # Make the state-transition ode matrix if it hasn't already been made
        if self.stm_ode_func is None:
            self.stm_ode_func = self.make_stmode(deriv_func, y0g.shape[0])

        if sol.arcs is None:
            sol.arcs = [(0, len(solinit.t)-1)]

        arc_seq = sol.aux['arc_seq']
        num_arcs = len(arc_seq)
        if len(sol.arcs) % 2 == 0:
            raise Exception('Number of arcs must be odd!')

        left_idx, right_idx = map(np.array, zip(*sol.arcs))
        ya = sol.y[:, left_idx]
        yb = sol.y[:, right_idx]

        tmp = np.arange(num_arcs+1, dtype=np.float32)*sol.t[-1]
        tspan_list = [(a, b) for a, b in zip(tmp[:-1], tmp[1:])]

        if solinit.parameters is None:
            nParams = 0
        else:
            nParams = solinit.parameters.size

        # Initial state of STM is an identity matrix with an additional column of zeros per parameter
        stm0 = np.hstack((np.eye(nOdes), np.zeros((nOdes,nParams)))).reshape(nOdes*(nOdes+nParams))
        n_iter = 1  # Initialize iteration counter
        converged = False  # Convergence flag

        # Ref: Solving Nonlinear Equations with Newton's Method By C. T. Kelley # TODO: Reference this in the docstring
        # Global Convergence and Armijo's Rule, pg. 11
        alpha = 1
        beta = 1
        r0 = None
        prop = Propagator(**self.ivp_args)
        y0stm = np.zeros((len(stm0)+nOdes))
        yb = np.zeros_like(ya)
        try:
            while True:
                phi_list = []
                phi_full_list = []
                t_list = []
                y_list = []
                with timeout(seconds=5000):
                    for arc_idx, tspan in enumerate(tspan_list):
                        y0stm[:nOdes] = ya[:, arc_idx]
                        y0stm[nOdes:] = stm0[:]
                        q0 = []
                        sol_ivp = prop(self.stm_ode_func, None, tspan, y0stm, q0, paramGuess, sol.aux, arc_idx)
                        t = sol_ivp.t
                        yy = sol_ivp.y.T
                        y_list.append(yy[:nOdes, :])
                        t_list.append(t)
                        yb[:, arc_idx] = yy[:nOdes, -1]
                        phi_full = np.reshape(yy[nOdes:, :].T, (len(t), nOdes, nOdes+nParams))
                        phi_full_list.append(np.copy(phi_full))
                        phi = np.reshape(yy[nOdes:, -1].T, (nOdes, nOdes+nParams))  # STM
                        phi_list.append(np.copy(phi))
                if n_iter == 1:
                    if not self.saved_code:
                        self.save_code()
                        self.saved_code = True

                # Break cycle if it exceeds the max number of iterations
                if n_iter > self.max_iterations:
                    logging.warning("Maximum iterations exceeded!")
                    break

                # Determine the error vector
                res = bc_func(ya, yb, paramGuess, sol.aux)

                # Break cycle if there are any NaNs in our error vector
                if any(np.isnan(res)):
                    print(res)
                    raise RuntimeError("Nan in residual")

                r1 = max(abs(res))
                if self.verbose:
                    logging.debug('Residual: ' + str(r1))

                if r1 > self.max_error:
                    raise RuntimeError('Error exceeded max_error')

                # Solution converged if BCs are satisfied to tolerance
                if r1 <= self.tolerance and n_iter > 1:
                    if self.verbose:
                        logging.info("Converged in "+str(n_iter)+" iterations.")
                    converged = True
                    break

                # Compute Jacobian of boundary conditions
                nBCs = len(res)
                J = self._bc_jac_multi(nBCs, phi_full_list, y_list, paramGuess, sol.aux, bc_func)

                # Compute correction vector
                if r0 is not None:
                    # if r1/r0 > 2:
                    #     logging.error('Residue increased more than 2x in one iteration!')
                    #     break
                    # beta = (r0-r1)/(alpha*r0)
                    beta = 1
                    if beta < 0:
                        beta = 1

                if r1 > 1:
                    alpha = 1/(2*r1)
                else:
                    alpha = 1

                r0 = r1

                # No damping if error within one order of magnitude of tolerance
                if r1 < min(10*self.tolerance, 1e-3):
                    alpha, beta = 1, 1

                try:
                    dy0 = alpha*beta*np.linalg.solve(J, -res)
                except:
                    dy0, *_ = np.linalg.lstsq(J, -res)
                    dy0 = alpha*beta*dy0

                # Apply corrections to states and parameters (if any)
                d_ya = np.reshape(dy0[:nOdes*num_arcs], (nOdes, num_arcs), order='F')
                if nParams > 0:
                    dp = dy0[nOdes*num_arcs:]
                    paramGuess += dp
                    ya = ya + d_ya
                else:
                    ya = ya + d_ya

                n_iter += 1
                logging.debug('Iteration #' + str(n_iter))

        except Exception as e:
            logging.warning(e)
            import traceback
            traceback.print_exc()

        if converged:
            sol.arcs = []
            timestep_ctr = 0
            for arc_idx, tt in enumerate(t_list):
                sol.arcs.append((timestep_ctr, timestep_ctr+len(tt)-1))

            sol.t = np.hstack(t_list)
            sol.y = np.column_stack(y_list)
            sol.parameters = paramGuess

        else:
            # Return a copy of the original guess if the problem fails to converge
            sol = copy.deepcopy(solinit)

        sol.converged = converged
        return sol



from beluga.visualization.renderers import BaseRenderer

import toyplot.browser, toyplot.qt, toyplot.html, toyplot.pdf, toyplot.png, toyplot.svg
from toyplot import *

class ToyPlot(BaseRenderer):
    def __init__(self, backend = 'browser'):
        self._figures = []
        backends = {'browser': toyplot.browser,
                    'pdf': toyplot.pdf,
                    'png': toyplot.png,
                    'svg': toyplot.svg,
                    'html': toyplot.html}
        self.backend = toyplot.browser

    def _get_figure(self,f):
        """
        Returns Figure instance from internal list using index

        Raises:
            ValueError if invalid index is used
        """
        try:
            fh =  self._figures[f]
            if fh is None:
                raise ValueError('Invalid figure handle specified!')
            return fh
        except:
            raise ValueError('Invalid figure handle specified!')

    def create_figure(self):
        """
        Creates a new figure and returns a handle (index into array)
        """
        canvas = toyplot.Canvas(width=600, height=600)
        self._figures.append(canvas)
        return len(self._figures)-1

    def close_figure(self,f):
        """
        Closes a specified figure
        """
        pass
        # close(self._get_figure(f))

    def show_figure(self,f):
        """
        Shows a specified figure
        """
        self.backend.show(self._get_figure(f))

    def show_all(self):
        """
        Show all rendered figures
        """
        for f in self._figures:
            self.backend.show(f)

    def render_plot(self,f,p):
        """
        Adds a line plot using the given data to the specified figure
        """
        canvas = self._get_figure(f);
        axes = canvas.axes()
        legends_data = []
        # has_legend = False
        for line in p.plot_data:
            # has_legend = has_legend or (line['label'] is not None)
            for dataset in line['data']:
                plt = axes.plot(dataset['x_data'], dataset['y_data'])
            # TODO: Legends for line_series may not work properly
            if line['label'] is not None:
                legends_data.append((line['label'],plt))
        # canvas.legend(legends_data);

        if p._xlabel is not None:
            axes.x.label.text = p._xlabel
        if p._ylabel is not None:
            axes.y.label.text = p._ylabel
        if p._title is not None:
            axes.label.text = p._title

    def render_subplot(self,f,index,plot):
        """
        Adds a subplot to the specified figure
        """
        pass

if __name__ == '__main__':
    from beluga.visualization.elements import Plot
    import dill

    r = ToyPlot()
    fig = r.create_figure()

    with open('/Users/tantony/dev/tantony-beluga/examples/planarHypersonic/phu_2k5_eps2.dill','rb') as f:
        out = dill.load(f)

    p = Plot(-1,-1)
    p.line('v/1000','h/1000')
    p.xlabel('v (km/s)')
    p.ylabel('h (km)')
    p.title('Altitude vs. Velocity')
    p.preprocess(out['solution'],out['problem_data'])

    r.render_plot(fig,p)
    r.show_figure(fig)

    fig = r.create_figure()
    p = Plot(-1,-1)
    p.line('theta*re/1000','h/1000')
    p.xlabel('downrange (km)')
    p.ylabel('altitude (km)')
    p.title('Altitude vs. Downrange')
    p.preprocess(out['solution'],out['problem_data'])
    r.render_plot(fig,p)
    r.show_figure(fig)



from .optimlib import *
import sympy as sym
from beluga.utils import sympify2, keyboard
import simplepipe as sp
from beluga.problem import SymVar
import logging
import numpy as np


def make_control_dae(states, costates, controls, constraints, dhdu, xi_init_vals, guess, derivative_fn):
    """
    Make's control law for dae (ICRM) formulation.

    :param states:
    :param costates:
    :param controls:
    :param constraints:
    :param dhdu:
    :param xi_init_vals:
    :param guess:
    :param derivative_fn:
    :return:
    """
    equality_constraints = constraints.get('equality', [])
    if len(equality_constraints) > 0:
        mu_vars = [sympify('mu'+str(i+1)) for i in range(len(equality_constraints))]
        mu_lhs = [sympify(c.expr) for c in equality_constraints]
    else:
        mu_vars = mu_lhs = []

    g = dhdu + mu_lhs
    X = [state for state in states] + [costate for costate in costates]
    U = [c for c in controls] + mu_vars

    xdot = sym.Matrix([sympify(state.eom) for state in states] + [sympify(lam.eom) for lam in costates])
    # Compute Jacobian
    dgdX = sym.Matrix([[derivative_fn(g_i, x_i) for x_i in X] for g_i in g])
    dgdU = sym.Matrix([[derivative_fn(g_i, u_i) for u_i in U] for g_i in g])

    udot = dgdU.LUsolve(-dgdX*xdot) # dgdU * udot + dgdX * xdot = 0

    dae_states = U
    dae_equations = list(udot)
    dae_bc = g

    if guess.start is not None:
        guess.start.extend(xi_init_vals)
    guess.dae_num_states = len(U)

    yield mu_vars
    yield mu_lhs
    yield dae_states
    yield dae_equations
    yield dae_bc
    yield guess
    yield dgdX
    yield dgdU


def generate_problem_data(workspace):
    """Generates the `problem_data` dictionary used for code generation."""

    tf_var = sympify('tf')

    dgdX = []
    for i, row in enumerate(workspace['dgdX'].tolist()):
        for j, expr in enumerate(row):
            if expr != 0:
                dgdX.append(f'dgdX[{i},{j}] = {expr}')

    dgdU = []
    for i, row in enumerate(workspace['dgdU'].tolist()):
        for j, expr in enumerate(row):
            if expr != 0:
                dgdU.append(f'dgdU[{i},{j}] = {expr}')

    problem_data = {
        'method': 'icrm',
        'problem_name': workspace['problem_name'],
        'aux_list': [
            {
            'type' : 'const',
            'vars': [str(k) for k in workspace['constants']]
            }
        ],
        'state_list':
            [str(x) for x in it.chain(workspace['states'], workspace['costates'])]
        ,
        'parameter_list': [str(tf_var)] + [str(p) for p in workspace['parameters']],
        'x_deriv_list': [str(tf_var*state.eom) for state in workspace['states']],
        'lam_deriv_list':[str(tf_var*costate.eom) for costate in workspace['costates']],
        'deriv_list':
            [str(tf_var*state.eom) for state in workspace['states']] +
            [str(tf_var*costate.eom) for costate in workspace['costates']]
        ,
        'states': workspace['states'],
        'costates': workspace['costates'],
        'constants': workspace['constants'],
        'parameters': [tf_var] + workspace['parameters'],
        'controls': workspace['controls'],
        'mu_vars': workspace['mu_vars'],
        'quantity_vars': workspace['quantity_vars'],
        'dae_var_list':
            [str(dae_state) for dae_state in workspace['dae_states']],
        'dae_eom_list':
            ['(tf)*('+str(dae_eom)+')' for dae_eom in workspace['dae_equations']],
        'dae_var_num': len(workspace['dae_states']),

        'costate_eoms': [ {'eom':[str(_.eom*tf_var) for _ in workspace['costates']], 'arctype':0} ],
        'ham': workspace['ham'],

        's_list': [],
        'bc_list': [],
        'num_states': 2*len(workspace['states']),
        'num_params': len(workspace['parameters']) + 1,
        'dHdu': [str(_) for _ in it.chain(workspace['dhdu'], workspace['mu_lhs'])],
        'bc_initial': [str(_) for _ in workspace['bc_initial']],
        'bc_terminal': [str(_) for _ in it.chain(workspace['bc_terminal'], workspace['dae_bc'])],
        'num_bc': len(workspace['bc_initial'])+len(workspace['bc_terminal'])+ len(workspace['dae_bc']),
        'control_options': [],
        'control_list': [str(u) for u in workspace['controls']+workspace['mu_vars']],
        'num_controls': len(workspace['controls'])+len(workspace['mu_vars']),
        'ham_expr': str(workspace['ham']),
        'quantity_list': workspace['quantity_list'],
        'bc_free_mask': workspace['bc_free_mask'],
        'dgdX': dgdX,
        'dgdU': dgdU,
        'nOdes': 2*len(workspace['states']) + len(workspace['dae_states']),
    }

    return problem_data


# Implement workflow using simplepipe and functions defined above
ICRM = sp.Workflow()
ICRM.add_task(BaseWorkflow)
ICRM.add_task(make_control_dae,
            inputs=('states', 'costates', 'controls', 'constraints', 'dhdu', 'xi_init_vals', 'guess', 'derivative_fn'),
            outputs=('mu_vars', 'mu_lhs', 'dae_states', 'dae_equations', 'dae_bc', 'guess', 'dgdX', 'dgdU'))
ICRM.add_task(generate_problem_data, inputs='*', outputs=('problem_data'))



import dill
import numpy as np
import numexpr as ne
import collections
from sympy import *
from sympy.utilities.lambdify import lambdify
from beluga.utils import keyboard

class Plot(object):
    """
    Represents a single plot with axes, labels, expressions to evaluate etc.
    """
    def __init__(self, step, sol, mesh_size, datasource, colormap=None):
        self.step_index = step
        self.sol_index = sol
        self.mesh_size = mesh_size
        self.datasource = datasource
        self.plot_data = []
        self._title = self._xlabel = self._ylabel = None
        self.postprocess_list = []
        self.colormap = colormap

    def xlabel(self, label):
        self._xlabel = label
        return self

    def ylabel(self, label):
        self._ylabel = label
        return self

    def step(self, _step):
        self.step_index = _step
        return self

    def solution(self, _sol):
        self.sol_index = _sol
        return self

    def title(self, title_txt):
        """
        Sets the title of the figure
        """
        self._title = title_txt
        return self

    def line(self, x_expr, y_expr, label=None, style=None, step = None, sol = None, datasource = None):
        """
        Adds a new line plot to the figure
        """
        if datasource is None:
            datasource = self.datasource
        if style is None:
            style = {}
        # TODO: Datatype sanity checks needed here
        self.plot_data.append({'type':'line', 'x':x_expr, 'y':y_expr, 'style':style, 'label':label, 'step':step, 'sol':sol, 'datasource': datasource})
        return self

    def line_series(self, x_expr, y_expr, label=None, style=None, step = None, start = 0, skip = 0, end = -1, datasource = None):
        if datasource is None:
            datasource = self.datasource
        if style is None:
            style = {}
        self.plot_data.append({'type':'line_series', 'x':x_expr, 'y':y_expr, 'style':style, 'label':label, 'step':step, 'start':start, 'skip':skip, 'end': end, 'datasource': datasource})
        return self

    def line3d(self, x_expr, y_expr, z_expr, label=None, style=None, step = None, sol = None, datasource = None):
        if datasource is None:
            datasource = self.datasource
        if style is None:
            style = {}
        # TODO: Datatype sanity checks needed here
        self.plot_data.append({'type':'line3d', 'x':x_expr, 'y':y_expr, 'z':z_expr, 'style':style, 'label':label, 'step':step, 'sol':sol, 'datasource': datasource})
        return self

    def preprocess(self):
        """
        Evaluates the expressions using the supplied data
        """
        for line in self.plot_data:
            solution = line['datasource'].get_solution()
            problem_data  = line['datasource'].get_problem()

            step_idx = line['step'] if line['step'] is not None else self.step_index
            line['data'] = []
            if line['type'] == 'line' or line['type'] == 'line3d':
                sol_idx = line['sol'] if line['sol'] is not None else self.sol_index
                sol = solution[step_idx][sol_idx]

                sol.prepare(problem_data, mesh_size=self.mesh_size, overwrite=True)
                line['data'].append({'x_data': sol.evaluate(line['x']),
                                     'y_data': sol.evaluate(line['y']),
                                     'z_data': sol.evaluate(line.get('z','0'))})
            elif line['type'] == 'line_series':
                sol_set = solution[step_idx]
                line['end'] = len(sol_set) if line['end'] == -1 else line['end']
                for ind in range(line['start'], line['end'], line['skip']+1):
                    sol = sol_set[ind]
                    sol.prepare(problem_data, mesh_size=self.mesh_size, overwrite=True)
                    # sol.prepare(problem_data)
                    line['data'].append({'x_data': sol.evaluate(line['x']),
                                         'y_data': sol.evaluate(line['y'])})
            else:
                raise ValueError('Invalid plot type specified')

    def postprocess(self, fn):
        self.postprocess_list.append(fn)
        return self

class PlotList(list):
    """
    A collection class for storing plot objects
    """
    def add_plot(self, plot=None):
        if plot is None:
            plot = Plot()
        self.append(plot)
        return plot



import os
import sys
import inspect
import logging

def root():
    """Returns beluga installation path."""
    return os.path.dirname(__file__)


def init_logging(logging_level, display_level, logfile):
    """Initializes the logging system"""
    # Define custom formatter class that formats messages based on level
    # Ref: http://stackoverflow.com/a/8349076/538379
    class InfoFormatter(logging.Formatter):
        """Custom logging formatter to output info messages by themselves"""
        info_fmt = '%(message)s'
        def format(self, record):
            # Save the original format configured by the user
            # when the logger formatter was instantiated
            format_orig = self._fmt

            # Replace the original format with one customized by logging level
            if record.levelno == logging.INFO:
                self._fmt = self.info_fmt
                # For Python>3.2
                self._style = logging.PercentStyle(self._fmt)

            # Call the original formatter class to do the grunt work
            result = logging.Formatter.format(self, record)

            # Restore the original format configured by the user
            self._fmt = format_orig
            # For Python>3.2
            self._style = logging.PercentStyle(self._fmt)

            return result

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    fh = logging.FileHandler(logfile)
    fh.setLevel(logging_level)

    # Set default format string based on logging level
    # TODO: Change this to use logging configuration file?
    if logging_level == logging.DEBUG:
        formatter = logging.Formatter('[%(levelname)s] %(asctime)s-%(module)s#%(lineno)d-%(funcName)s(): %(message)s')
    else:
        formatter = logging.Formatter('[%(levelname)s] %(asctime)s-%(filename)s:%(lineno)d: %(message)s')

    # Create logging handler for console output
    ch = logging.StreamHandler(sys.stdout)
    # Set console logging level and formatter
    ch.setLevel(display_level)
    formatter = InfoFormatter('%(filename)s:%(lineno)d: %(message)s')
    ch.setFormatter(formatter)

    logger.addHandler(fh)
    logger.addHandler(ch)



# def test_mcpi1():
#     """Test mcpi() against analytical solution"""
#     k  = [-0.5, -0.2]
#     def odefn(t,x,p,aux=None):
#         return np.array([k[0]*x[0],k[1]*x[1]])
#
#     y0 = np.array([10,-50])
#     tspan = np.array([0, 1.0])
#     [t1, x1] = mcpi(odefn, tspan, y0, N=10, xtol=1e-5)
#     x1_expected = np.array([y*np.exp(k_*t1) for (y,k_) in zip(y0,k)]).T
#     print(x1)
#     print(x1_expected[-1])
#     assert (x1 - x1_expected < 1e-5).all()



from beluga.visualization.renderers import BaseRenderer
from matplotlib.pyplot import *
import matplotlib.colors as colors
import matplotlib.cm as cmx
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np

class MatPlotLib(BaseRenderer):
    """
    A renderer class that implements matplotlib
    """
    def __init__(self):
        self._figures = []

    def _get_figure(self,f):
        """
        Returns Figure instance from internal list using index

        Raises:
            ValueError if invalid index is used
        """
        try:
            fh =  self._figures[f]
            if fh is None:
                raise ValueError('Invalid figure handle specified!')
            return fh
        except:
            raise ValueError('Invalid figure handle specified!')

    def create_figure(self):
        """
        Creates a new figure and returns a handle
        """
        self._figures.append(figure())
        return len(self._figures)-1

    def close_figure(self,f):
        """
        Closes a specified figure
        """
        close(self._get_figure(f))

    def show_figure(self,f,block=False):
        """
        Shows a specified figure
        """
        show(self._get_figure(f))

    def show_all(self):
        """
        Show all rendered figures
        """
        show()

    def render_plot(self,f,p):
        """
        Adds a line plot using the given data to the specified figure
        """
        fh = self._get_figure(f);
        has_legend = False
        num_lines = sum(1 for line in p.plot_data for d in line['data'] )
        if p.colormap is not None:
            cm_subsection = np.linspace(0.0, 1.0, num_lines)
            colors = [ p.colormap(x) for x in cm_subsection ]
        else:
            colors = [None]*num_lines

        i = 0
        ax3d = None
        for line in p.plot_data:
            has_legend = has_legend or (line['label'] is not None)
            override_color = None
            if isinstance(line['style'], dict):
                override_color = line['style'].pop('color', None)

            if line['type'] == 'line3d':
                fig = self._get_figure(f)
                if ax3d is None:
                    ax3d = fig.add_subplot(111, projection='3d')

            for dataset in line['data']:
                # Allow overriding colormap fom style vector
                if override_color is not None:
                    line_color = override_color
                else:
                    line_color = colors[i]

                if line['type'] == 'line3d':
                    if(len(dataset['x_data'])!=len(dataset['y_data'])):
                        continue

                    if isinstance(line['style'], str):
                        ax3d.plot3D(dataset['x_data'],dataset['y_data'],dataset['z_data'],line['style'],label=line['label'],figure=fh, color=line_color,)
                    elif isinstance(line['style'], dict):
                        ax3d.plot3D(dataset['x_data'],dataset['y_data'],dataset['z_data'],label=line['label'],figure=fh, color=line_color,**line['style'])
                else:
                    if(len(dataset['x_data'])!=len(dataset['y_data'])):
                        continue

                    if isinstance(line['style'], str):
                        plot(dataset['x_data'],dataset['y_data'],line['style'],label=line['label'],figure=fh, color=line_color,)
                    elif isinstance(line['style'], dict):
                        plot(dataset['x_data'],dataset['y_data'],label=line['label'],figure=fh, color=line_color,**line['style'])
                i += 1
        if has_legend:
            fh.gca().legend()
        if p._xlabel is not None:
            xlabel(p._xlabel,figure=fh)
        if p._ylabel is not None:
            ylabel(p._ylabel,figure=fh)
        if p._title is not None:
            title(p._title,figure=fh)
        grid(True)

    def render_subplot(self,f,index,plot):
        """
        Adds a subplot to the specified figure
        """
        pass

if __name__ == '__main__':
    from beluga.visualization.elements import Plot
    import dill

    r = MatPlotLibRenderer()
    fig = r.create_figure()

    with open('/Users/tantony/dev/mjgrant-beluga/examples/data.dill','rb') as f:
        out = dill.load(f)

    p = Plot(0,-1)
    p.x('v/1000')
    p.y('h/1000')
    p.xlabel('v (km/s)')
    p.ylabel('h (km)')
    p.title('Altitude vs. Velocity')
    p.preprocess(out['solution'],out['problem_data'])

    r.render_plot(fig,p)
    r.show_figure(fig)



import numba
import functools as ft
import simplepipe as sp
import logging
from sympy.utilities.lambdify import lambdastr
import pystache
import imp
import sympy as sym
import itertools as it
# from beluga.bvpsol.algorithms import Shooting
import beluga
import sys
import numpy as np
from math import *
from beluga.utils import keyboard


def load_eqn_template(problem_data, template_file, renderer = pystache.Renderer(escape=lambda u: u)):
    """
    Loads pystache template and uses it to generate code.

    Parameters
    ----------
        problem_data - dict
            Workspace defining variables for template

        template_file - str
            Path to template file to be used

        renderer
            Renderer used to convert template file to code

    Returns
    -------
    Code generated from template
    """
    template_path = beluga.root()+'/optimlib/templates/'+problem_data['method']+'/'+template_file
    with open(template_path) as f:
        tmpl = f.read()
        # Render the template using the data
        code = renderer.render(tmpl, problem_data)
        return code


def create_module(problem_data):
    """
    Creates a new module for storing compiled code.

    Parameters
    ----------
    problem_data - dict
        Problem data dictionary

    Returns
    -------
    New module for holding compiled code
    """
    problem_name = problem_data['problem_name']
    module = imp.new_module('_beluga_'+problem_name)



    # module.corner_fns = problem_data['corner_fns']
    # module.compute_hamiltonian = problem_data['ham_fn']
    # module.costate_eoms = problem_data['costate_eoms']

    return module


def make_control_and_ham_fn(control_opts, states, costates, parameters, constants, controls, mu_vars, quantity_vars, ham, constraint_name=None):
    controls = sym.Matrix([_._sym for _ in controls])
    constants = sym.Matrix([_._sym for _ in constants])
    states = sym.Matrix([_.name for _ in states])
    costates = sym.Matrix([_.name for _ in costates])
    parameters = sym.Matrix(parameters)
    tf_var = sym.sympify('tf')
    unknowns = list(it.chain(controls, mu_vars))

    ham_args = [*states, *costates, *parameters, *constants, *unknowns]
    u_args = [*states, *costates, *parameters, *constants]

    if constraint_name is not None:
        ham_args.append(constraint_name)
        u_args.append(constraint_name)
    else:
        ham_args.append('___dummy_arg___')
        u_args.append('___dummy_arg___')
    control_opt_mat = sym.Matrix([[option.get(u, '0')
                                    for u in unknowns]
                                    for option in control_opts])


    # control_opt_fn = sym.lambdify(u_args, control_opt_mat)
    # print('Making control fn with args',u_args)
    control_opt_fn = make_sympy_fn(u_args, control_opt_mat)

    # print('Making ham fn with args', ham_args)
    ham_fn = make_sympy_fn(ham_args, ham.subs(quantity_vars))

    num_unknowns = len(unknowns)
    num_options = len(control_opts)
    num_states = len(states)
    num_params = len(parameters)
    constraint_name = str(constraint_name)

    def compute_control_fn(t, X, p, aux):
        X = X[:(2*num_states+1)]
        C = aux['const'].values()
        p = p[:num_params]
        s_val = aux['constraint'].get((constraint_name, 1), -1)

        try:
            u_list = control_opt_fn(*X, *p, *C, s_val)
        except Exception as e:
            # print('oh nooes')
            # from beluga.utils import keyboard
            # keyboard()
            raise

            # keyboard()
        ham_val = np.zeros(num_options)
        for i in range(num_options):
            try:
                ham_val[i] = ham_fn(*X, *p, *C, *u_list[i], s_val)
            except:
                print(X, p, C, u_list[i])
                raise
        # if len(ham_val) == 0:
        #     keyboard()
        return u_list[np.argmin(ham_val)]

    yield compute_control_fn
    yield ham_fn


def make_functions(problem_data, module):

    unc_control_law = problem_data['control_options']
    states = problem_data['states']
    costates = problem_data['costates']
    parameters = problem_data['parameters']
    constants = problem_data['constants']
    controls = problem_data['controls']
    mu_vars = problem_data['mu_vars']
    quantity_vars = problem_data['quantity_vars']
    ham = problem_data['ham']

    logging.info('Making unconstrained control')
    control_fn, ham_fn = make_control_and_ham_fn(unc_control_law,states,costates,parameters,constants,controls,mu_vars,quantity_vars,ham)

    # problem_data['ham_fn'] = ham_fn
    module.ham_fn = ham_fn
    control_fns = [control_fn] # Also makethiss
    logging.info('Processing constraints')
    for arc_type, s in enumerate(problem_data['s_list'],1):
        u_fn, ham_fn = make_control_and_ham_fn(s['control_law'], states, costates, parameters, constants, controls, mu_vars, quantity_vars, s['ham'], s['name'])
        control_fns.append(u_fn)

    module.control_fns = control_fns
    module.ham_fn = ham_fn

    yield module
    yield control_fns[0]


def compile_code_py(code_string, module, function_name):
    """
    Compiles a function specified by template in filename and stores it in
    self.compiled

    Parameters
    ----------
    code_string - str
        String containing the python code to be compiled

    module - dict
        Module in which the new functions will be defined

    function_name - str
        Name of the function being compiled (this must be defined in the
        template with the same name)

    Returns:
        Module for compiled function
        Compiled function
    """
    # For security
    module.__dict__.update({'__builtin__': {}})
    exec(code_string, module.__dict__)
    return getattr(module, function_name, None)


def make_njit_fn(args, fn_expr):
    fn_str = lambdastr(args, fn_expr).replace('MutableDenseMatrix', '')\
                                                  .replace('(([[', '[') \
                                                  .replace(']]))', ']')
    jit_fn = numba.njit(parallel=True, nopython=True)(eval(fn_str))
    return jit_fn


def make_sympy_fn(args, fn_expr):
    if hasattr(fn_expr, 'shape'):
        output_shape = fn_expr.shape
    else:
        output_shape = None

    if output_shape is not None:
        jit_fns = [make_njit_fn(args, expr) for expr in fn_expr]
        len_output = len(fn_expr)

        # @numba.jit(parallel=True, nopython=True)
        def vector_fn(*args):
            output = np.zeros(output_shape)
            for i in numba.prange(len_output):
                output.flat[i] = jit_fns[i](*args)
            return output
        return vector_fn
    else:
        return make_njit_fn(args, fn_expr)



from .BaseRenderer import BaseRenderer
from .MatPlotLib import MatPlotLib
# from .ToyPlot import ToyPlot
from .Bokeh import Bokeh

import os
import glob
modules = glob.glob(os.path.dirname(__file__)+"/*.py")
__all__ = [ os.path.basename(f)[:-3] for f in modules]



import numpy as np
from numba import vectorize
from functools import lru_cache
from math import *

@vectorize(['float64(float64, float64)'], nopython=True, target='cpu')
def absdiff(a, b):
    return abs(a-b)

@lru_cache(maxsize=None)
def mcpi_init(N):
    """
    Given Chebyshev series of order N
    Returns tau, C_x and C_a matrices used in MCPI
    """
    tau = np.fromiter((cheby_node(i, N) for i in range(N+1)), np.float64)
    T = np.array([[cheby_poly(tau_j, k) for k in range(N+1)] for tau_j in tau])   # Matrix used for computation
    W = np.diag(np.ones(N+1))
    W[0,0] /= 2
    C_x = T @ W  # x = C_x * alpha for coefficients alpha

    V = np.diag((2/N)*np.ones(N+1))
    V[0,0] /= 2
    V[N, N] /= 2

    R = np.diag([1] + [1/(2*r) for r in range(1, N+1)])
    S = np.diag(np.ones(N), k=-1) + np.diag(-np.ones(N), k=1)
    S[0,2:N] = np.fromiter(((-1)**(r+1)*(1/(r-1) - 1/(r+1)) for r in range(2, N)), np.float64)  ## r = 2, 3, ..., N-1
    S[0, :2] = [1, -1/2]
    S[0, -1] = (-1)**(N+1)/(N-1)

    C_a = R @ S @ T @ V # For each variable, Beta = x_0 + C_a * g()

    return (tau, C_x, C_a)

@lru_cache(maxsize=None)
def cheby_node(j, N):
    """
    Evaluates CGL-node tau_j for chebyshev series of order 'N'
    """
    return cos(j*pi/N)


@lru_cache(maxsize=None)
def cheby_poly(tau_j, k):
    """
    Evaluates chebyshev polynomial of order 'k' at point tau_j
    """
    return cos(k*acos(tau_j))

def cheby_eval(beta, tspan):
    """
    Evaluates chebyshev series given by coefficients 'beta' over the given mesh 'tspan'
    """
    N = len(beta)-1
    tau_fine = np.linspace(-1, 1, len(tspan))
    T = np.array([[cheby_poly(tau_j, k) for tau_j in tau_fine] for k in range(N+1)])
    W = np.diag(np.ones(N+1))
    W[0,0] /= 2
    x_new = (W @ T).T @ beta
    return x_new

def mcpi(ode_fn, tspan, x_0, N=51, maxiter=501, xtol=1e-4, args=()):
    """
    Propagates equations in EOM with initial conditions in x_0
    Returns:
        tt: Mesh points for the solution
        x_new: Solution (each column = time history of one state)
    """
    #TODO: Use numexpr and strings to speed up EOMs
    x_0 = np.array(x_0, dtype=np.float64)
    w1 = (tspan[-1]-tspan[0])/2
    w2 = (tspan[-1]+tspan[0])/2

    tau, C_x, C_a = mcpi_init(N)      # tau is in reverse order (1 to -1)
    C_a = w1 * C_a
    t_arr = tau * w1 + w2

    x_guess = np.tile(x_0, [N+1, 1])  # Each column -> time history of one state
    x0_twice = 2*x_guess[0]
    g = np.empty_like(x_guess)

    for ctr in range(maxiter):
        # Evaluate vectorized EOM over full time domain
        ode_fn(t_arr, x_guess, g, *args)

        beta = C_a @ g
        beta[0]+= x0_twice          # Compute Chebyshev coefficients of solution
        x_new = C_x @ beta          # Compute solution

        err1 = np.max(absdiff(x_new, x_guess))
        if (err1 < xtol):
#             print('Converged in %d iterations.' % ctr)
            break
        x_guess = x_new

    if len(tspan) > 2:
        # If tspan is bigger than 2 elements, evaluate solution on that mesh
        return tspan, cheby_eval(beta, tspan)
    else:
        # Else return solution on CGL mesh
        # return t_arr, np.array(list(reversed(x_new)))
        return t_arr[::-1], np.flipud(x_new)
        # return t_arr[::-1], np.flipud(x_new)



from beluga.ivpsol.ivpsol import Propagator, Algorithm, Trajectory, reconstruct, integrate_quads

import os
import glob
modules = glob.glob(os.path.dirname(__file__)+"/*.py")
__all__ = [ os.path.basename(f)[:-3] for f in modules]



import pytest
from beluga.optimlib import init_workspace
from beluga.optimlib import make_augmented_params, make_augmented_cost
from beluga.problem import SymVar

def test_init_workspace():
    class emptyobj(object):
        def __new__(cls):
            obj = super(emptyobj, cls).__new__(cls)
            obj.dae_num_states = 0
            return obj

    guess = emptyobj()
    from beluga.problem import OCP
    problem = OCP('test_problem')

    # Throw an error with no independent variable defined.
    with pytest.raises(KeyError):
        init_workspace(problem, guess)

    problem.independent('t', 's')
    problem.state('x', 'v', 'm')
    problem.state('v', 'g + u', 'm/s')
    problem.control('u', 'N')
    problem.constant('g', 9.80665, 'm/s^2')
    problem.path_cost('1', 's')
    problem.constraints().initial('x-x_0', 'm')
    problem.constraints().initial('v-v_0', 'm/s')
    problem.constraints().terminal('x-x_f', 'm')
    problem.scale(m='x', s='x/v', N=1)

    ws = init_workspace(problem, guess)

    assert isinstance(ws, dict)
    assert ws['problem_name'] == 'test_problem'


def test_make_augmented_cost():
    cost = SymVar({'expr': '1', 'unit': 'nd'}, sym_key='expr')
    constraints = dict()
    constraints['initial'] = [{'expr': 'x-x_0', 'unit': 'nd'}, {'expr': 'y-y_0', 'unit': 'nd'}]
    constraints['terminal'] = [{'expr': 'x-x_f', 'unit': 'nd'}]
    constraints = {c_type: [SymVar(c_obj, sym_key='expr') for c_obj in c_list]
                   for c_type, c_list in constraints.items()
                   if c_type != 'path'}

    constraints_adjoined = False
    location = 'initial'

    lambdas = make_augmented_params(constraints, constraints_adjoined, location)
    result = make_augmented_cost(cost, constraints, constraints_adjoined, location)
    assert result == 1
    assert len(lambdas) == 0

    constraints_adjoined = True

    lambdas = make_augmented_params(constraints, constraints_adjoined, location)
    result = make_augmented_cost(cost, constraints, constraints_adjoined, location)
    assert result == lambdas[0]*constraints['initial'][0] + lambdas[1]*constraints['initial'][1] + cost
    assert len(lambdas) == 2

    constraints_adjoined = False
    location = 'terminal'

    lambdas = make_augmented_params(constraints, constraints_adjoined, location)
    result = make_augmented_cost(cost, constraints, constraints_adjoined, location)
    assert result == 1
    assert len(lambdas) == 0

    constraints_adjoined = True

    lambdas = make_augmented_params(constraints, constraints_adjoined, location)
    result = make_augmented_cost(cost, constraints, constraints_adjoined, location)
    assert result == lambdas[0] * constraints['terminal'][0] + cost
    assert len(lambdas) == 1



import numpy as np
import numexpr as ne
from scipy.interpolate import InterpolatedUnivariateSpline
import math
from beluga.ivpsol import Trajectory


class Solution(Trajectory):
    def __new__(cls, *args, **kwargs):
        """
        Creates a new Solution object.

        :param args:
        :param kwargs:
        :return:
        """

        t = kwargs.get('t', None)
        y = kwargs.get('y', None)
        q = kwargs.get('q', None)
        u = kwargs.get('u', None)
        parameters = kwargs.get('parameters', None)
        aux = kwargs.get('aux', None)
        state_list = kwargs.get('state_list', None)
        arcs = kwargs.get('arcs', None)

        if t is not None:
            t = np.array(t, dtype=np.float64)
        if y is not None:
            y = np.array(y, dtype=np.float64)
        if q is not None:
            q = np.array(q, dtype=np.float64)
        if u is not None:
            u = np.array(u, dtype=np.float64)

        obj = super(Solution, cls).__new__(cls, t, y, q, u)

        if parameters is not None:
            obj.parameters = np.array(parameters, dtype=np.float64)
        else:
            obj.parameters = np.array([])

        if aux is None:
            obj.aux = {"initial": [], "terminal": [], "const": {}, "parameters": [], "arc_seq": (0,)}
        else:
            obj.aux = aux

        obj.state_list = state_list
        obj.var_dict = None
        obj.converged = False
        obj.arcs = arcs

        obj.y_splines = None
        obj.u_splines = None
        obj.extra = None
        obj.converged = False
        return obj

    # TODO: Remove this and use Trajectory()'s interpolation.
    def init_interpolate(self):
        """
        Fits splines to all states in the solution data
        """
        self.y_splines = []
        self.u_splines = []
        for arc_start, arc_end in self.arcs:
            y_spline_arc = []
            u_spline_arc = []
            for j,row in enumerate(self.y):
                spline = InterpolatedUnivariateSpline(self.t[arc_start:arc_end+1], row[arc_start:arc_end+1])
                y_spline_arc.append(spline)

            if len(self.u.shape) ==1 :
                spline = InterpolatedUnivariateSpline(self.t[arc_start:arc_end+1], self.u[arc_start:arc_end+1])
                u_spline_arc.append(spline)
            else:
                for j,row in enumerate(self.u):
                    spline = InterpolatedUnivariateSpline(self.t[arc_start:arc_end+1], row[arc_start:arc_end+1])
                    u_spline_arc.append(spline)

            self.y_splines.append(y_spline_arc)
            self.u_splines.append(u_spline_arc)

    def interpolate(self, new_t, new_arcs, overwrite=False):
        """
        Interpolates solution data over a new mesh of 't'

        new_t : new mesh to evaluate
        overwrite: Should the current solution be overwritten?
        """
        # Account for old data files with no sol_splines
        if self.y_splines is None or self.u_splines is None:
            self.init_interpolate()

        new_y = np.hstack([np.vstack([spline(new_t[arc_start:arc_end+1]) for spline in spline_arc])
                           for (arc_start, arc_end), spline_arc in zip(new_arcs, self.y_splines)])
        new_u = np.hstack([np.vstack([spline(new_t[arc_start:arc_end+1]) for spline in spline_arc])
                           for (arc_start, arc_end), spline_arc in zip(new_arcs, self.u_splines)])

        if overwrite:
            self.t = new_t
            self.y = new_y
            self.u = new_u
            self.arcs = new_arcs

        return new_y, new_u

    def prepare(self, problem_data, mesh_size=None, overwrite=False):
        """
        Creates the dictionary required to evaluate expressions over the solution

        mesh_size: Evaluate over new mesh
        overwrite: Overwrite existing solution with new mesh
        """

        if not hasattr(self, 'arcs'):
            self.arcs = ((0,len(self.t)-1),)
        if mesh_size is not None and mesh_size > len(self.t)*len(self.arcs):
            # Update solution to use new mesh if needed
            new_t_list = []
            new_arcs = []
            arc_ctr = 0
            for arc_start, arc_end in self.arcs:
                new_t_list.append(np.linspace(self.t[arc_start], self.t[arc_end], mesh_size))
                new_arcs.append((arc_ctr, arc_ctr + mesh_size - 1))
                arc_ctr += mesh_size

            new_t = np.hstack(new_t_list)
            (new_y, new_u) = self.interpolate(new_t, new_arcs, overwrite=overwrite)
        else:
            new_t, new_y, new_u, new_arcs = self.t, self.y, self.u, self.arcs
        t, y, u, p, arcs = self.t, self.y, self.u, self.parameters, self.arcs
        if not overwrite:
            t, y, u, p, arcs = new_t, new_y, new_u, new_arcs

        #TODO: Write test for prepare()
        #TODO: Make state_list a part of the Solution object

        # Define every aux variable (such as constants) in the dictionary
        variables = [(aux_name, np.ones_like(t)*aux_val)
                     for aux_type in self.aux
                     if isinstance(self.aux[aux_type], dict)
                     for (aux_name, aux_val) in self.aux[aux_type].items()
                     if isinstance(aux_val, float)
                     ]
        # Define state variables
        # Have to do in this order to override state values with arrays
        variables += [(state,np.array(y[idx,:], dtype=np.float64))
                       for idx,state in enumerate(problem_data['state_list'])]

        # Define control variables
        variables += [(control,np.array(u[idx,:], dtype=np.float64))
                       for idx,control in enumerate(problem_data['control_list'])]


        # Define constants
        # variables += [(str(constant),np.ones_like(x)*float(self.aux['const'][].value))
        #                for idx,constant in enumerate(problem_data['constants'])]

        t_list = []
        tf_ind = problem_data['parameter_list'].index('tf')
        last_t = 0
        for arc_start, arc_end in arcs:
            start_t = t[arc_start]
            # from beluga.utils import keyboard
            # keyboard()
            t_list.append(last_t+(t[arc_start:arc_end+1] - start_t))
            last_t += y[tf_ind, arc_start]

        variables += [('t', np.hstack(t_list))]

        if 'quantity_list' in problem_data:
            self.qvars = {q['name']:q['expr'] for q in problem_data['quantity_list']}
        else:
            self.qvars = problem_data.get('quantity_vars', {})

        qvars = []
        for q_k, q_v in self.qvars.items():
            qvars.append((str(q_k), ne.evaluate(str(q_v), dict(variables+qvars))))

        variables += qvars
        self.var_dict = dict(variables)
        self.var_dict['pi'] = math.pi

    def evaluate(self,expr):
        """
        Evaluates an expression or custom function involving the variables in this solution

        The caller is responsible for calling the prepare() method first
        """

        if callable(expr):
            # Custom plot function
            return expr(**{arg_name:self.var_dict[arg_name] for arg_name in inspect.getargspec(expr).args})
        else:
            # String specified in plot
            return ne.evaluate(expr,self.var_dict)



from beluga.ivpsol import Propagator
import numpy as np
from math import *

def compute_hamiltonian(t,X,p,aux,u):
    [x,y,v,lamX,lamY,lamV,tf] = X[:7]
    g = aux['const']['g']

    thetta = u[0]
    return lamX*v*cos(thetta) + g*lamV*sin(thetta) + lamY*v*sin(thetta) + 1


def compute_control(t,X,p,aux):
    [x,y,v,lamX,lamY,lamV,tf] = X[:7]
    g = aux['const']['g']

    thetta_saved = float('inf')
    ham_saved = float('inf')

    try:
        thetta = -acos(-((lamX*v)/sqrt(g**2*lamV**2+2*g*lamV*lamY*v+(lamX**2+lamY**2)*v**2)))
    except:
        thetta = 0
    ham = compute_hamiltonian(t,X,p,aux,[thetta])
    if ham < ham_saved:
        ham_saved = ham
        thetta_saved = thetta
    try:
        thetta = acos(-((lamX*v)/sqrt(g**2*lamV**2+2*g*lamV*lamY*v+(lamX**2+lamY**2)*v**2)))
    except:
        thetta = 0
    ham = compute_hamiltonian(t,X,p,aux,[thetta])
    if ham < ham_saved:
        ham_saved = ham
        thetta_saved = thetta

    try:
        thetta = -acos((lamX*v)/sqrt(g**2*lamV**2+2*g*lamV*lamY*v+(lamX**2+lamY**2)*v**2))
    except:
        thetta = 0
    ham = compute_hamiltonian(t,X,p,aux,[thetta])
    if ham < ham_saved:
        ham_saved = ham
        thetta_saved = thetta

    try:
        thetta = acos((lamX*v)/sqrt(g**2*lamV**2+2*g*lamV*lamY*v+(lamX**2+lamY**2)*v**2))
    except:
        thetta = 0
    ham = compute_hamiltonian(t,X,p,aux,[thetta])
    if ham < ham_saved:
        ham_saved = ham
        thetta_saved = thetta

    if thetta_saved == float('inf'):
        thetta_saved = 0
    return thetta_saved


def brachisto_ode(t,_X,_p,aux):
    [x, y, v, lamX, lamY, lamV, tf] = _X[:7]
    g = aux['const']['g']

    thetta = compute_control(t, _X, _p, aux)
    xdot = tf*np.array([v*cos(thetta),
                        v*sin(thetta),
                        g*sin(thetta),
                        0,
                        0,
                        -(lamX*cos(thetta) + lamY*sin(thetta)),
                        0])
    return xdot


inputs = [np.array([0.0, 0.0, 1.0, -0.1, -0.1, -0.1, 0.1]),
          [np.array([0.0, 0.0, 1.0, -0.1, -0.1, -0.1, 0.1]), np.array([0.0, 0.0, 1.0, -0.1, -0.1, -0.1, 0.1])]
          ]

x_end = np.array([2.57442406e-02, -1.46160497e-01, 1.96663900e+00, -1.00000000e-01,
                  -1.00000000e-01, -1.08178965e-01, 1.00000000e-01])


def test_ode45_2():
    x0 = inputs[0]
    q0 = []
    tspan = np.array([0, 1.0])
    prop = Propagator()
    aux = {'const': {'g': -9.81}}
    solout = prop(brachisto_ode, None, tspan, x0, q0, [], aux)
    x1 = solout.y

    assert (x1[-1,:] - x_end < 1e-5).all()


from setuptools import find_packages, setup
import os, sys

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open('README.rst') as file:
    long_description = file.read()

modules = ['beluga.bvpsol',
           'beluga.bvpsol.algorithms',
           'beluga.codegen',
           'beluga.continuation',
           'beluga.continuation.strategies',
           'beluga.ivpsol',
           'beluga.ivpsol.integrators',
           'beluga.optimlib',
           'beluga.utils',
           'beluga.visualization',
           'beluga.visualization.renderers']

tests = ['beluga.bvpsol.tests',
         'beluga.ivpsol.tests',
         'beluga.ivpsol.integrators.tests',
         'beluga.optimlib.tests']

setup(name="beluga",
      version="0.1.6",
      description="An indirect trajectory optimization framework.",
      long_description=long_description,
      author="Michael Sparapany",
      author_email='msparapa@purdue.edu',
      platforms=["any"],  # or more specific, e.g. "win32", "cygwin", "osx"
      license="MIT",
      url="https://github.com/Rapid-Design-of-Systems-Laboratory/beluga",
      py_modules=['beluga'],
      packages=['beluga'] + modules + tests,
    #   scripts=['bin/beluga'],
      entry_points={
          'console_scripts': [
              'beluga = beluga.__main__:main'
          ]
      },
      install_requires=requirements,
      include_package_data = True
      )



from beluga.bvpsol.algorithms import Shooting
from beluga.bvpsol import Solution
import numpy as np

tol = 1e-3

def test_Shooting_1():
    # Full 2PBVP test problem
    # This is the simplest BVP

    def odefun(t, X, p, const, arc):
        return np.hstack((X[1], -abs(X[0])))

    def bcfun(X0,Xf,p,aux):
        return np.hstack((X0[0], Xf[0]+2))

    algo = Shooting()
    solinit = Solution()
    solinit.t = np.linspace(0,4,2)
    solinit.y = np.array([[0,1],[0,1]]).T
    out = algo.solve(odefun, None, bcfun, solinit)
    assert out.y.T[0][0] < tol
    assert out.y.T[0][1] - 2.06641646 < tol
    assert out.y.T[-1][0] + 2 < tol
    assert out.y.T[-1][1] + 2.87588998 < tol
    assert out.t[-1] - 4 < tol
    assert abs(out.y.T[0][1] - solinit.y.T[0][1]) > tol
    assert abs(out.y.T[-1][0] - solinit.y.T[-1][0]) - 2 < tol

def test_Shooting_2():
    # Full 2PBVP test problem
    # This is calculating the 4th eigenvalue of Mathieu's Equation
    # This problem contains an adjustable parameter.

    def odefun(t, X, p, const, arc):
        return np.hstack((X[1], -(p[0] - 2 * 5 * np.cos(2 * t)) * X[0]))

    def bcfun(X0, Xf, p, aux):
        return np.hstack((X0[1], Xf[1], X0[0] - 1))

    algo = Shooting()
    solinit = Solution()
    solinit.t = np.linspace(0, np.pi, 30)
    solinit.y = np.vstack((np.cos(4 * solinit.t), -4 * np.sin(4 * solinit.t)))
    solinit.parameters = np.array([15])

    out = algo.solve(odefun, None, bcfun, solinit)
    assert out.t[-1] - np.pi < tol
    assert out.y.T[0][0] - 1 < tol
    assert out.y.T[0][1] < tol
    assert out.y.T[-1][0] - 1 < tol
    assert out.y.T[-1][1] < tol
    assert out.parameters[0] - 17.09646175 < tol

def test_Shooting_3():
    # This problem contains a parameter, but it is not explicit in the BCs.
    # Since time is buried in the ODEs, this tests if the BVP solver calculates
    # sensitivities with respect to parameters.
    def odefun(t, X, p, const, arc):
        return (1 * p[0])

    def bcfun(X0, Xf, p, aux):
        return np.hstack((X0[0] - 0, Xf[0] - 2))

    algo = Shooting()
    solinit = Solution()
    solinit.t = np.linspace(0, 1, 2)
    solinit.y = np.array([[0], [0]]).T
    solinit.parameters = np.array([1])
    out = algo.solve(odefun, None, bcfun, solinit)
    assert out.parameters - 2 < tol



# from beluga.optim import *
from optimalcontrol.elements import Problem
from sympy import *
from sympy.core.function import AppliedUndef, Function
import pystache, imp, inspect, logging, os
import re as _re

import beluga.bvpsol.BVP as BVP

from beluga.utils import sympify, keyboard, ipsh
from beluga.optim.problem import *
import dill
import numpy as np
from collections import Iterable

class OCDAENumerical(object):
    name = 'dae-numerical'

    def get_actions(self):
        # Tuple: (func, arg1, arg2, ...)
        self.actions = [
            (self.process_quantities),
            (self.make_costates),
            (self.make_aug_cost, 'initial'),
            (self.make_aug_cost, 'terminal'),
            # (self.process_path_constraints, problem),
        ]
        return self.actions

    def run(self):
        for action in self.actions:
            if isinstance(action, Iterable):
                action[0](*action[1:])
            else:
                action()

        # pystache renderer without HTML escapes
        renderer = pystache.Renderer(escape=lambda u: u)

    def __init__(self, problem, cached=True):
        """!
        \brief     Initializes all of the relevant necessary conditions of opimality.
        \author    Michael Grant
        \author    Thomas Antony
        \version   0.2
        \date      05/15/16
        """

        self.problem = problem
        self.get_actions()

        self.aug_cost = {}
        self.costates = []
        self.costate_rates = []
        self.ham = sympify('0')
        self.ham_ctrl_partial = []
        self.ctrl_free = []
        self.parameter_list = []
        self.bc_initial = []
        self.bc_terminal = []
        # from .. import Beluga # helps prevent cyclic imports
        self.compile_list = ['deriv_func','bc_func','compute_control']
        # self.template_prefix = Beluga.config.getroot()+'/beluga/bvpsol/templates/'
        self.template_suffix = '.py.mu'
        self.dae_states = []
        self.dae_equations = []

    def cache_bvp(self, problem, filename=None):
        """
        \brief Saves BVP object into file on disk
        Arguments:
            problem : Problem object
            filename: Full path to cache file (optional)
                      default value: <self.problem.name>_bvp.dat
        \date  01/27/2016
        """
        if filename is None:
            filename = problem.name+'_bvp.dat'

        with open(filename,'wb') as f:
            try:
                logging.info('Caching BVP information to file')
                bvp = dill.dump(self.bvp,f)
                return True
            except:
                logging.warn('Failed to save BVP to '+filename)
                return False

    def load_bvp(self, problem, filename=None):
        """
        \brief  Loads pre-computed BVP object from cache file
        \author Thomas Antony
        Arguments:
            problem : Problem object
            filename: Full path to cache file (optional)
                      default value: <self.problem.name>_bvp.dat
        \date  01/27/2016
        """
        if filename is None:
            filename = problem.name+'_bvp.dat'

        if not os.path.exists(filename):
            return None

        with open(filename,'rb') as f:
            try:
                logging.info('Loading BVP information from cache')
                bvp = dill.load(f)
                return bvp
            except Exception as e:
                logging.warn('Failed to load BVP from '+filename)
                logging.debug(e)
                return None

    def derivative(self, expr, var, dependent_variables):
        """
        Take derivative taking pre-defined quantities into consideration

        dependent_variables: Dictionary containing dependent variables as keys and
                             their expressions as values
        """
        dep_var_names = dependent_variables.keys()
        dep_var_expr = [(expr) for (_,expr) in dependent_variables.items()]

        dFdq = [diff(expr, dep_var).subs(dependent_variables.items()) for dep_var in dep_var_names]
        dqdx = [diff(qexpr, var) for qexpr in dep_var_expr]

        # Chain rule + total derivative
        out = sum(d1*d2 for d1,d2 in zip(dFdq, dqdx)) + diff(expr, var)
        return out

    def make_costate_rate(self, states):
        """!
        \brief     Creates the symbolic differential equations for the costates.
        \author    Michael Grant
        \author    Thomas Antony
        \version   0.1
        \date      06/30/15
        """

        # TODO: Automate partial derivatives of numerical functions
        # for state in states:
        #     rate = diff(sympify('-1*(' + self.ham + ')'),state)
        #     # numerical_diff = rate.atoms(Derivative)
        #     self.costate_rates.append(str(rate))
        self.costate_rates = [self.derivative(-1*(self.ham),state, self.quantity_vars) for state in states]
        # self.costate_rates.append(str(diff(sympify(
        # '-1*(' + self.ham + ')'),state)))

        for state in states:
            corate=self.derivative(-1*(self.ham),state,self.quantity_vars)
            custom_diff=corate.atoms(Derivative)
            repl = [] #Replacements for custom functions
            for d in custom_diff:
                for f in d.atoms(AppliedUndef): #Just extracts (never more than 1 f)
                    for v in f.atoms(Symbol):
                        if str(v)==str(state): #This should only happen once!
                            replv=sympify('im('+str(f.subs(v,v+_h))+')/1e-30')
                            repl.append((d,replv))
            self.costate_rates.append(corate.subs(repl))
        self.costate_rates=sympify2(self.costate_rates)

        #h = 1e-30
        #j = symbols('1j')

        #self.costate_rates = []

        #for state in states:
        #    state = sympify(state)
        #    H = self.make_state_dep_ham(state, self.problem)
        #    if H != 0:
        #        self.costate_rates.append('im(' + str((H.subs(state, (state + h * j)))) + ')/' + str(h))
        #    else:
        #        self.costate_rates.append('0')
        #self.costate_rates=sympify2(self.costate_rates)
        #print(Matrix([sympify2(state.process_eqn) for state in self.problem.states()]+ self.costate_rates))
        #quit()

    def make_costate_rate_numeric(self, states):
        """!
        \brief     Creates the symbolic differential equations for the costates.
        \author    Michael Grant
        \author    Sean Nolan
        \version   0.1
        \date      06/22/16
        """

        h = 1e-100
        j = symbols('1j')

        self.costate_rates = []

        for state in states:
            state = sympify(state)
            H = self.make_state_dep_ham(state, self.problem)
            if H != 0:
                self.costate_rates.append('-np.imag(' + str((H.subs(state, (state + h * j)))) + ')/' + str(h))
            else:
                self.costate_rates.append('0')


    def make_ctrl_partial(self, controls):
        """!
        \brief     Symbolically compute dH/du where H is the Hamiltonian and u is the control.
        \author    Michael Grant
        \author    Thomas Antony
        \version   0.1
        \date      06/30/15
        """

        self.ham_ctrl_partial = []
        _h = sympify2('1j*(1e-30)')

        for ctrl in controls:
            dHdu = self.derivative(sympify(self.ham), ctrl, self.quantity_vars)
            custom_diff = dHdu.atoms(Derivative)
            for d in custom_diff:
                for f in d.atoms(AppliedUndef):
                    for v in f.atoms(Symbol):
                        if str(v)==str(ctrl):
                            replv=sympify('im('+str(f.subs(v,v+_h))+')/1e-30')
                            repl.append((d,replv))

            self.ham_ctrl_partial.append(dHdu.subs(repl))
            # Substitute "Derivative" with complex step derivative
            #repl = {(d,im(f.func(v+1j*1e-30))/1e-30) for d in custom_diff
            #            for f,v in zip(d.atoms(AppliedUndef),d.atoms(Symbol))}

            #self.ham_ctrl_partial.append(dHdu.subs(repl))

    def selective_expand(self, expr, var_select, subs_list):
        # Expands the expressions specified by subs_list if they contain the
        # variables in var_select
        select_subs = [qty for qty in subs_list.items() for var in var_select if var.sym in qty[1].atoms()]
        # Substitute only relevant symbols in the expression
        return expr.subs(select_subs)

    def make_ctrl(self, problem, mode='dae'):
        if mode == 'dae':
            self.make_ctrl_dae(problem)
        else:
            self.make_ctrl_analytic(problem.controls())

    def make_ctrl_dae(self, problem):
        """!
        \brief     Compute EOMs for controls.
        \author    Michael Grant
        \author    Thomas Antony
        \version   0.1
        \date      04/10/16
        """
        if len(self.equality_constraints) > 0:
            self.mu_vars = [sympify('mu'+str(i+1)) for i in range(len(self.equality_constraints))]
            self.mu_lhs = [sympify(c.expr) for c in self.equality_constraints]
        else:
            self.mu_vars = self.mu_lhs = []

        g = self.ham_ctrl_partial + self.mu_lhs
        X = [state.sym for state in problem.states()] + [Symbol(costate) for costate in self.costates]
        U = [c.sym for c in problem.controls()] + self.mu_vars

        xdot = Matrix([sympify(state.process_eqn) for state in problem.states()] + self.costate_rates)
        # Compute Jacobian
        dgdX = Matrix([[self.derivative(g_i, x_i, self.quantity_vars) for x_i in X] for g_i in g])
        dgdU = Matrix([[self.derivative(g_i, u_i, self.quantity_vars) for u_i in U] for g_i in g])

        udot = dgdU.LUsolve(-dgdX*xdot); # dgdU * udot + dgdX * xdot = 0

        self.dae_states = U

        self.dae_equations = list(udot)
        self.dae_bc = g


    def make_ctrl_analytic(self, controls):
        """!
        \brief     Symbolically compute the solutions for the control along control-unconstrained arcs.
        \author    Michael Grant
        \author    Thomas Antony
        \version   0.1
        \date      06/30/15
        """
        # Solve all controls simultaneously
        logging.info("Finding optimal control law ...")

        # If equality constraints are present
        # We need to solve for 'mu's as well
        lhs = self.ham_ctrl_partial
        vars = [c.sym for c in controls]
        self.mu_vars = []
        self.mu_lhs = []

        if len(self.equality_constraints) > 0:
            self.mu_vars = [sympify('mu'+str(i+1)) for i in range(len(self.equality_constraints))]

            self.mu_lhs = [sympify(c.expr) for c in self.equality_constraints]
        try:
            var_list = list(vars + self.mu_vars)
            # var_list = vars
            eqn_list = []
            for eqn in list(lhs + self.mu_lhs ):
                eqn_list.append(self.selective_expand(eqn, controls, self.quantity_vars))
            logging.info("Attempting using SymPy ...")
            logging.debug("dHdu = "+str(eqn_list))

            ctrl_sol = solve(eqn_list, var_list, dict=True)
            #TODO: ^ uncomment once done with the aircraft noise problem
            #ctrl_sol = CtrlSols() #I'm going to hell for this
            #logging.info("Grabbed the control laws for the aircraft noise!")

            # raise ValueError() # Force mathematica
        except ValueError as e:  # FIXME: Use right exception name here
            logging.debug(e)
            logging.info("No control law found")
            from beluga.utils.pythematica import mathematica_solve
            logging.info("Attempting using Mathematica ...")
            #var_sol = mathematica_solve(lhs+self.mu_lhs,vars+self.mu_vars)
            # TODO: Extend numerical control laws to mu's
            ctrl_sol = var_sol
            if ctrl_sol == []:
                logging.info("No analytic control law found, switching to numerical method")

        #logging.info("Done")
        # solve() returns answer in the form
        # [ {ctrl1: expr11, ctrl2:expr22},
        #   {ctrl1: expr21, ctrl2:expr22}]
        # Convert this to format required by template
        self.control_options = [ [{'name':str(ctrl), 'expr':str(expr)}
                                    for (ctrl,expr) in option.items()]
                                for option in ctrl_sol]

    def make_aug_cost(self, location):
        """!
        \brief     Symbolically create the augmented cost functional.
        \author    Michael Grant
        \author    Thomas Antony
        \version   0.1
        \date      06/30/15
        """
        aug_cost   = self.problem.cost[location].expr
        constraint = self.problem.constraints(location)
        # Do in two steps so that indices are "right"
        # TODO: apply quantities
        filtered_list = [c for c in constraint if c.type==location]
        self.parameter_list += [c.make_multiplier(ind) for (ind,c) in enumerate(filtered_list,1)]
        # self.aug_cost[location] = aug_cost + ''.join(' + (%s)' % c.make_aug_cost(ind)
        #                         for (ind,c) in enumerate(filtered_list,1))

        self.aug_cost[location] = aug_cost + sum([c.make_aug_cost(ind)
                                                 for (ind,c) in enumerate(filtered_list,1)])

    def make_costate_bc(self, states, location):
        """!
        \brief     Symbolically create the boundary conditions at initial and final locations.
        \author    Michael Grant
        \author    Thomas Antony
        \version   0.1
        \date      06/30/15
        """

        if location is 'initial':
            sign = sympify('-1')
        elif location is 'terminal':
            sign = sympify('1')

        cost_expr = sign * (self.aug_cost[location])

        #TODO: Fix hardcoded if conditions
        #TODO: Change to symbolic
        if location == 'initial':
            # Using list comprehension instead of loops
            # lagrange_ changed to l. Removed hardcoded prefix
            self.bc_initial += [str(sympify(state.make_costate()) - self.derivative(sympify(cost_expr),state.sym, self.quantity_vars))
                                    for state in states]
        else:
            # Using list comprehension instead of loops
            self.bc_terminal += [str(sympify(state.make_costate()) - self.derivative(sympify(cost_expr),state.sym,self.quantity_vars))
                                    for state in states]

    def make_ham(self, problem):
        """!
        \brief     Symbolically create the Hamiltonian.
        \author    Michael Grant
        \author    Thomas Antony
        \version   0.1
        \date      06/30/15
        """
        #TODO: Make symbolic
        self.ham = sympify(problem.cost['path'].expr)
        for i in range(len(problem.states())):
            self.ham += sympify(self.costates[i]) * (sympify(problem.states()[i].process_eqn))

        # Adjoin equality constraints
        for i in range(len(self.equality_constraints)):
            self.ham += sympify('mu'+str(i+1)) * (sympify(self.equality_constraints[i].expr))

    def make_state_dep_ham(self, state, problem):
        """!
        \brief     Symbolically create the Hamiltonian.
        \author    Michael Grant
        \author    Sean Nolan
        \version   0.1
        \date      06/22/16
        """

        H = sympify(0)

        new_terms = sympify2(problem.cost['path'].expr)
        atoms = new_terms.atoms()
        if state in atoms:
            H += new_terms


        for i in range(len(problem.states())):
            new_terms = sympify2(self.costates[i]) * (sympify2(problem.states()[i].process_eqn))
            atoms = new_terms.atoms()
            if state in atoms:
                H += new_terms

        # Adjoin equality constraints
        for i in range(len(self.equality_constraints)):
            new_terms = sympify2('mu' + str(i + 1)) * (sympify2(self.equality_constraints[i].expr))
            atoms = new_terms.atoms()
            if state in atoms:
                H += new_terms

        raw_terms = H.as_terms()
        H_parts = [raw_terms[0][k][0] for k in range(len(raw_terms[0]))]
        new_H = sympify(0)
        for part in H_parts:
            store = sympify(0)
            tries = 0
            while (part != store) & (tries <= 5):
                store = part
                part = part.expand()
                tries += 1
            raw_terms = part.as_terms()
            sub_parts = [raw_terms[0][k][0] for k in range(len(raw_terms[0]))]
            new_part = sympify(0)
            for sub_part in sub_parts:
                atoms = sub_part.atoms()
                if state in atoms:
                    # sub_part = sub_part.factor()
                    new_part += sub_part
            new_part = new_part.factor()
            if len(str(new_part)) < 50:
                 new_part = new_part.simplify(ratio=1.0)
            new_H += new_part

        # print('New:' + str(len(str(new_H))))
        # print('Old:' + str(len(str(H))))

        return new_H

    # Compiles a function template file into a function object
    # using the given data
    def compile_function(self,filename,verbose=False):
        """
        Compiles a function specified by template in filename and stores it in
        self.compiled

        Returns:
            bool: True if successful

        Raises:
            ValueError: If 'problem_data' or 'compiled' is not defined
        """
        with open(filename) as f:
            tmpl = f.read()

            if self.problem_data is None:
                raise ValueError('Problem data not defined. Unable to compile function.')

            if self.compiled is None:
                raise ValueError('Problem module not defined. Unable to compile function.')

            # Render the template using the data
            code = self.renderer.render(tmpl,self.problem_data)
            # if verbose and 'compute_control' in filename:
            # if verbose:
            logging.debug(code)

            # For security
            self.compiled.__dict__.update({'__builtin__':{}})
            return exec(code,self.compiled.__dict__)

    # TODO: Maybe change all constraint limits (initial, terminal etc.) to be 'constants' that can be changed by continuation?
    def sanitize_constraint(self,constraint,problem):
        """
        Checks the initial/terminal constraint expression for invalid symbols
        Also updates the constraint expression to reflect what would be in code
        """
        if constraint.type == 'initial':
            pattern = r'([\w\d\_]+)_0'
            prefix = '_x0'
        elif constraint.type == 'terminal':
            pattern = r'([\w\d\_]+)_f'
            prefix = '_xf'
        else:
            raise ValueError('Invalid constraint type')

        m = _re.findall(pattern,constraint.expr)
        invalid = [x for x in m if x not in problem.states()]

        if not all(x is None for x in invalid):
            raise ValueError('Invalid expression(s) in boundary constraint:\n'+str([x for x in invalid if x is not None]))

        # Create new variable for output to avoid mutating original object
        output = Constraint()
        output.type = constraint.type
        output.unit = constraint.unit
        output.expr = _re.sub(pattern,prefix+r"['\1']",constraint.expr)

        return output

    # def process_systems(self,problem):
    #     """Traverses dynamic systems list and extracts information"""
    #     for (system_type,system_list) in problem.systems.items():
    #         for idx,system_inst in enumerate(system_list):
    #             # new_states = [state.add_prefix(system_type+'_'+str(idx)+'_')
    #             #                 for state in system_inst.states]
    #             new_states = [state
    #                             for state in system_inst.states]
    #     # print(new_states)

    # 1/4 s (c1-c2) = slope at x=0
    # when slope = 1, s = 4/(c1-c2)
    # wyen slope = 10, s = 40/(c1-c2)
    def get_satfn(self, var, ubound=None, lbound=None, slopeAtZero=1):
        # var -> varible inside saturation function
        if ubound is None and lbound is None:
            raise ValueError('At least one bound should be specified for the constraint.')
        if ubound == lbound:
            raise ValueError('The upper and lower bounds should be different.')

        # Symmetric bounds
        if ubound is None:
            # Return lower bound sat fn
            #ubound = -lbound
            return lbound + exp(var)
        elif lbound is None:
            # Return upper bound sat fn
            #lbound = -ubound
            return ubound - exp(-var)
        else:
            print(ubound)
            s = 4*slopeAtZero/(ubound - lbound)
            return ubound - ( ubound - lbound )/( 1 + exp(s*var) )

    def process_path_constraints(self, problem):
        constraints = problem.constraints().get('path')
        quantity_subs = self.quantity_vars.items()

        path_cost_expr = sympify(problem.cost['path'].expr)
        path_cost_unit = sympify(problem.cost['path'].unit)
        if path_cost_expr == 0:
            logging.debug('No path cost specified, using unit from terminal cost function')
            problem.cost['path'].unit = problem.cost['terminal'].unit
            path_cost_unit = sympify(problem.cost['terminal'].unit)

        logging.debug('Path cost is of unit: '+str(path_cost_unit))
        time_unit = Symbol('s')

        for (ind,c) in enumerate(constraints):
            # Determine order of constraint
            logging.debug('Processing path constraint: '+c.label)
            order = 0
            cq = [sympify(c.expr)]
            dxdt = [sympify(state.process_eqn) for state in problem.states()]

            # Zeroth order constraints have no 'xi' state
            xi_vars = []
            h = []
            while True:
                control_found = False

                for u in problem.controls():
                    if u.sym in cq[-1].subs(quantity_subs).atoms():
                        logging.info('Constraint is of order '+str(order))
                        control_found = True
                        break
                if control_found:
                    break

                dcdx = [self.derivative(cq[-1], state.sym, self.quantity_vars) for state in problem.states()]

                # Chain rule (Assume there is no explciit time-dependence) to find dcdt
                cq.append(sum(d1*d2 for d1,d2 in zip(dcdx, dxdt)))
                order = order + 1

                # Create the auxiliary state variables
                xi_vars.append(Symbol('xi'+str(ind+1)+str(order)))

            # Create the smoothing control variable
            xi_vars.append(Symbol('ue'+str(ind+1)))

            # TODO: Fix constraint object to accept two limits
            c_limit = sympify(c.limit)
            if c_limit.is_Number:
                # TODO: Allow continuation on constraints
                # Define new hidden constant
                c_limit = sympify('_'+c.label)
                print(c.limit)
                problem.constant(str(c_limit),float(c.limit),c.unit)
                logging.debug('Added constant '+str(c_limit))

            if c.direction == '>':
                c.lbound = c_limit
                c.ubound = -c_limit
            elif c.direction == '<':
                c.ubound = c_limit
                c.lbound = -c_limit
            else:
                raise ValueError('Invalid direction specified for constraint')

            psi = self.get_satfn(xi_vars[0], ubound=c.ubound, lbound=c.lbound, slopeAtZero=1)
            psi_vars = [(Symbol('psi'+str(ind+1)), psi)]

            # Add to quantity list
            self.quantity_vars[Symbol('psi'+str(ind+1))] = psi
            self.quantity_list.append({'name':('psi'+str(ind+1)), 'expr':str(psi)})

            # m-th order constraint needs up to m-th derivative of psi to be defined
            psi_i = psi
            for i in range(order):
                psi_i = diff(psi_i, xi_vars[0])
                # psi_vars.append((Symbol('psi'+str(ind+1)+str(i+1)+'('+str(xi_vars[0])+')'), psi_i))
                psi_vars.append((Symbol('psi'+str(ind+1)+str(i+1)), psi_i))
                self.quantity_vars[Symbol('psi'+str(ind+1)+str(i+1))] =  psi_i
                self.quantity_list.append({'name':('psi'+str(ind+1)+str(i+1)), 'expr':str(psi_i)})

            # psi_vars = psi_vars + []
            psi_var_sub = [(v,k) for k,v in psi_vars]

            # FIXME: Hardcoded h derivatives for now
            # h = [psi_vars[0][0]]
            # h.append(psi_vars[1][0]*xi_vars[1]) # psi'*xi12
            # h.append(psi_vars[2][0]*xi_vars[1] + psi_vars[1][0]*xi_vars[2]) # psi''*xi12 + psi'*xi13
            # psi'''*xi12 + xi13*psi12'' + psi12*xi13 + psi11*ue1
            # h.append(psi_vars[3][0]*xi_vars[1] + 2 * psi_vars[2][0]*xi_vars[2] + psi_vars[1][0]*xi_vars[3] )

            #TODO: Hardcoded 't' as independent variable with unit of 's'
            # c_vals = [80e3, -5000, 9.539074102210087] # third number is vdot at zero approx
            c_vals = np.ones(order)*0.1
            h = [psi_vars[0][1]]
            for i in range(order):
                # Add 'xi' state
                problem.state(str(xi_vars[i]), str(xi_vars[i+1]),'('+c.unit+')/s^('+str(i)+')')
                # Constraint all cq at initial point (forms constraints for xi_ij)
                problem.constraints().initial(str(cq[i] - h[i]),'('+c.unit+')/s^('+str(i)+')')
                # Add to initial guess vector
                problem.guess.start.append(c_vals[i])

                dhdxi = [diff(h[i], xi_v) for xi_v in xi_vars[:-1]]
                dhdt  = sum(d1*d2 for d1,d2 in zip(dhdxi,xi_vars[1:])) # xi11dot = xi12 etc.
                dhdt = dhdt.subs(psi_var_sub)
                h.append(dhdt)

            # Add the smoothing control with the right unit
            ue_unit = sympify('('+c.unit+')/(s^('+str(order)+'))')
            problem.control(str(xi_vars[-1]), str(ue_unit))
            logging.debug('Adding control '+str(xi_vars[-1])+' with unit '+str(ue_unit))

            # Add equality constraint
            cqi_unit = ue_unit*time_unit
            problem.constraints().equality(str(cq[-1] - h[-1]),str(cqi_unit))

            # Add smoothing factor
            eps_const = Symbol('eps_'+c.label)
            eps_unit = (path_cost_unit/ue_unit**2)/time_unit #Unit of integrand
            problem.constant(str(eps_const), 1e-6, str(eps_unit))
            logging.debug('Adding smoothing factor '+str(eps_const)+' with unit '+str(eps_unit))

            # Append new control to path cost
            path_cost_expr = path_cost_expr + eps_const*xi_vars[-1]**2

        logging.debug('Updated path cost is: '+str(path_cost_expr))
        problem.cost['path'].expr = str(path_cost_expr)

        u_constraints = problem.constraints().get('control')

        for (ind,c) in enumerate(u_constraints):
            w_i = sympify('uw'+str(ind+1))
            psi = self.get_satfn(w_i, ubound=sympify(c.ubound), lbound = sympify(c.lbound))

            # Add the smoothing control
            problem.control(str(w_i), c.unit)

            # Add equality constraint
            csym = sympify(c.expr)
            problem.constraints().equality(str(csym - psi),c.unit)

            uw_unit = symipfy2(c.unit)
            eps_const = Symbol('eps_'+str(ind+1))
            eps_unit = (path_cost_unit/uw_unit**2)/time_unit #Unit of integrand
            problem.constant(str(eps_const), 1, str(eps_unit))

            # problem.state('costC2','eps2*('+str(w_i)+'^2)','m^2/s^2')
    def process_quantities(self,):
        # Should this be moved into __init__ ?
        # self.process_systems(problem)
        logging.info('Processing quantity expressions')
        # Process quantities
        # Substitute all quantities that show up in other quantities with their expressions
        # TODO: Sanitize quantity expressions
        # TODO: Check for circular references in quantity expressions
        if len(self.problem.quantity()) > 0:
            quantity_subs = [(qty.var, qty.expr) for qty in self.problem.quantity()]
            quantity_sym, quantity_expr = zip(*quantity_subs)
            quantity_expr = [qty_expr.subs(quantity_subs) for qty_expr in quantity_expr]

            # Use substituted expressions to recreate quantity expressions
            quantity_subs = [(qty_var,qty_expr) for qty_var, qty_expr in zip(quantity_sym, quantity_expr)]
            # Dictionary for use with mustache templating library
            self.quantity_list = [{'name':str(qty_var), 'expr':str(qty_expr)} for qty_var, qty_expr in zip(quantity_sym, quantity_expr)]

            # Dictionary for substitution
            self.quantity_vars = dict(quantity_subs)
        else:
            self.quantity_list = quantity_subs = []
            self.quantity_vars = {}

    def make_costates(self):
        self.costates = [state.make_costate() for state in self.problem.states()]

    def get_bvp(self,problem,mode='dae'):
        """Perform variational calculus calculations on optimal control problem
           and returns an object describing the boundary value problem to be solved

        Returns: bvpsol.BVP object
        """
        # Process intermediate variables
        self.process_quantities(problem)

        # Regularize path constraints using saturation functions
        self.process_path_constraints(problem)

        # self.state_subs = [(state.sym, sympify(state.process_eqn)) for state in problem.states()]
        ## Create costate list
        # self.costates = [state.make_costate() for state in problem.states()]
        self.make_costates(problem)

        # for i in range(len(self.problem.states())):
        #     self.costates.append(self.problem.states()[i].make_costate())

        # Build augmented cost strings
        # aug_cost_init = sympify(problem.cost['initial'].expr)
        self.make_aug_cost('initial')
        # self.make_aug_cost(aug_cost_init, problem.constraints(), 'initial')

        # aug_cost_term = sympify(problem.cost['terminal'].expr)
        self.make_aug_cost('terminal')
        # self.make_aug_cost(aug_cost_term, problem.constraints(), 'terminal')

        # Add state boundary conditions
        self.bc_initial = [self.sanitize_constraint(x,problem).expr
                            for x in problem.constraints().get('initial')]
        self.bc_terminal = [self.sanitize_constraint(x,problem).expr
                    for x in problem.constraints().get('terminal')]

        self.equality_constraints = problem.constraints().get('equality')

        ## Unconstrained arc calculations
        # Construct Hamiltonian
        self.make_ham(problem)
        logging.debug('Hamiltonian : '+str(self.ham))
        # Get list of all custom functions in the problem
        # TODO: Check in places other than the Hamiltonian?
        # TODO: Move to separate method?
        func_list = sympify(self.ham).atoms(AppliedUndef)

        # TODO: Change this to not be necessary
        self.problem = problem

        # Load required functions from the input file
        new_functions = {(str(f.func),getattr(problem.input_module,str(f.func)))
                            for f in func_list
                            if hasattr(problem.input_module,str(f.func)) and
                                inspect.isfunction(getattr(problem.input_module,str(f.func)))}

        problem.functions.update(new_functions)

        undefined_func = [f.func for f in func_list if str(f.func) not in problem.functions]

        if not all(x is None for x in undefined_func):
            raise ValueError('Invalid function(s) specified: '+str(undefined_func))

        # Compute costate conditions
        self.make_costate_bc(problem.states(),'initial')
        self.make_costate_bc(problem.states(),'terminal')

        # TODO: Make this more generalized free final time condition
        # HARDCODED tf variable
        time_constraints = problem.constraints().get('independent')
        if len(time_constraints) > 0:
            self.bc_terminal.append('tf - 1')
        else:
            # Add free final time boundary condition
            self.bc_terminal.append('_H - 0')

        # Compute costate process equations
        if mode == 'numerical':
            self.make_costate_rate_numeric(problem.states())
        else:
            self.make_costate_rate(problem.states())
        self.make_ctrl_partial(problem.controls())

        # # Add support for state and control constraints
        # problem.state('xi11','xi12','m')
        # problem.state('xi12','ue1','m')
        # self.costates += ['eta11','eta12']   # Costates for xi
        #
        # # Add new states to hamiltonian
        # h1_3 = '(psi12*xi12^2 + psi11*ue1)';  # xi12dot = ue1
        # c1_2 = 'u'
        # self.ham += sympify('eta11*xi12 + eta12*ue1')  #
        # self.ham += sympify('mu1 * ('+c1_2+' - '+h1_3')')
        #
        # # TODO: Compute these automatically
        # self.costate_rates += ['mu1*(xi12**2*psi1_3 + psi12*ue1)',
        #                        'mu1*(2*psi12*xi12)  - eta1']
        #
        # Compute unconstrained control law
        # (need to add singular arc and bang/bang smoothing, numerical solutions)
        self.make_ctrl(problem, mode)

        # Create problem dictionary
        # NEED TO ADD BOUNDARY CONDITIONS

        # bc1 = [self.sanitize_constraint(x) for x in initial_bc]

        self.problem_data = {
        'aux_list': [
                {
                'type' : 'const',
                'vars': [const.var for const in problem.constants()]
                },
                {
                'type' : 'constraint',
                'vars': []
                },
                {
                'type' : 'function',
                'vars' : [func_name for func_name in problem.functions]
                }
         ],
         # TODO: Generalize 'tf' to independent variable for current arc
         'state_list':
             [str(state) for state in problem.states()] +
             [str(costate) for costate in self.costates] +
             ['tf']
         ,
         'parameter_list': [str(param) for param in self.parameter_list],
         'deriv_list':
             ['(tf)*(' + str(sympify(state.process_eqn)) + ')' for state in problem.states()] +
             ['(tf)*(' + str(costate_rate) + ')' for costate_rate in self.costate_rates] +
            #  ['(tf)*((' + str(costate_rate) + ').imag)' for costate_rate in self.costate_rates] +
             ['tf*0']   # TODO: Hardcoded 'tf'
         ,
         'state_rate_list':
            ['(tf)*(' + str(sympify2(state.process_eqn)) + ')' for state in problem.states()]
         ,
         'dae_var_list':
             [str(dae_state) for dae_state in self.dae_states],
         'dae_eom_list':
             ['(tf)*('+str(dae_eom)+')' for dae_eom in self.dae_equations],
         'dae_var_num': len(self.dae_states),
         'num_states': 2*len(problem.states()) + 1,
         'dHdu': [str(dHdu) for dHdu in self.ham_ctrl_partial] + self.mu_lhs,
         'left_bc_list': self.bc_initial + (self.dae_bc if (mode == 'dae') else []),
         'right_bc_list': self.bc_terminal,
         'control_options': [] if (mode == 'dae') else self.control_options,
         'control_list': [str(u) for u in problem.controls()] + [str(mu) for mu in self.mu_vars],
         'num_controls': len(problem.controls()) + len(self.mu_vars),  # Count mu multipliers
         'ham_expr':self.ham,
         # 'contr_dep_ham': [],
         'quantity_list': self.quantity_list,
        #  'dae_mode': mode == 'dae',
        }
    #    problem.constraints[i].expr for i in range(len(problem.constraints))

        # Create problem functions by importing from templates
        self.compiled = imp.new_module('_probobj_'+problem.name)

        if mode == 'dae':
            # self.template_suffix = '_dae' + self.template_suffix
            self.template_suffix = '_dae_num' + self.template_suffix
        if mode == 'num':
            self.template_suffix = '_num' + self.template_suffix

        compile_result = [self.compile_function(self.template_prefix+func+self.template_suffix, verbose=True)
                                        for func in self.compile_list]

        dhdu_fn = self.compiled.get_dhdu_func
        dae_num = len(problem.controls()) + len(self.mu_vars)

        self.bvp = BVP(self.compiled.deriv_func,self.compiled.bc_func,dae_func_gen=dhdu_fn,dae_num_states=dae_num)
        self.bvp.solution.aux['const'] = dict((const.var,const.val) for const in problem.constants())
        self.bvp.solution.aux['parameters'] = self.problem_data['parameter_list']
        self.bvp.solution.aux['function']  = problem.functions

        # TODO: Fix hardcoding of function handle name (may be needed for multivehicle/phases)?
        self.bvp.control_func = self.compiled.compute_control
        self.bvp.problem_data = self.problem_data
        # TODO: ^^ Do same for constraint values
        return self.bvp

if __name__ == '__main__':
    problem = Problem('brachisto')
    problem.quantity('foo','bar')
    im = OCDAENumerical(problem)
    im.run()
    print(im.quantity_list)



import abc
import logging
import collections as cl
import functools as ft
import simplepipe as sp
import numba
import sys
import pickle
from beluga.codegen import *
from beluga.utils import keyboard


class BaseAlgorithm(object):
    '''
    Object representing an algorithm that solves boundary valued problems.

    This object serves as a base class for other algorithms.
    '''
    # Define class as abstract class
    __metaclass__ = abc.ABCMeta

    # Define common interface for algorithm classes
    def __new__(cls, *args, **kwargs):
        obj = super(BaseAlgorithm, cls).__new__(cls)
        return obj

    @abc.abstractmethod
    def solve(self, deriv_func, bc_func, solinit):
        '''
        Method to solve the bvp with given arguments

        :param deriv_func: The ODE function.
        :param bc_func: The boundary conditions function.
        :param solinit: An initial guess for a solution to the BVP.
        :return: A solution to the BVP.
        '''
        raise NotImplementedError()

    def close(self):
        pass

    def preprocess(self, problem_data, use_numba=False):
        '''
        Code generation and compilation before running solver.

        :param problem_data:
        :param use_numba:
        :return: Code module.
        '''

        out_ws = PythonCodeGen({'problem_data': problem_data})
        logging.debug(out_ws['bc_func_code'])
        logging.debug(out_ws['deriv_func_code'])

        if use_numba:
            deriv_func = numba.njit(parallel=False, nopython=True)(out_ws['code_module'].deriv_func_nojit)
        else:
            deriv_func = out_ws['code_module'].deriv_func_nojit

        out_ws['deriv_func_fn'] = deriv_func
        out_ws['code_module'].deriv_func = deriv_func

        out_ws = out_ws

        bvp = BVP(out_ws['deriv_func_fn'], out_ws['bc_func_fn'], out_ws['compute_control_fn'])

        sys.modules['_beluga_' + problem_data['problem_name']] = out_ws['code_module']
        return out_ws['code_module'], bvp

    @staticmethod
    def load_code():
        logging.info('Loading compiled code ...')
        with open('codecache.pkl', 'rb') as f:
            bvp_data = pickle.load(f)

        return bvp_data

    def save_code(self):
        logging.info('Saving compiled code ...')
        bvp_data = {'deriv_fn': self.out_ws['code_module'].deriv_func}
        with open('codecache.pkl', 'wb') as f:
            pickle.dump(bvp_data, f, pickle.HIGHEST_PROTOCOL)

BVP = cl.namedtuple('BVP', 'deriv_func bc_func compute_control')

PythonCodeGen = sp.Workflow([
    # Create module for holding compiled code
    sp.Task(ft.partial(create_module), inputs='problem_data', outputs=('code_module')),

    sp.Task(make_functions, inputs=('problem_data', 'code_module'), outputs=('code_module','compute_control_fn')),
    # Load equation template files and generate code
    sp.Task(ft.partial(load_eqn_template,
            template_file='deriv_func.py.mu'),
            inputs='problem_data',
            outputs='deriv_func_code'),
    sp.Task(ft.partial(load_eqn_template,
            template_file='bc_func.py.mu'),
            inputs='problem_data',
            outputs='bc_func_code'),

    # Compile generated code
    sp.Task(ft.partial(compile_code_py, function_name='deriv_func'),
            inputs=['deriv_func_code', 'code_module'],
            outputs='deriv_func_fn'),
    sp.Task(ft.partial(compile_code_py, function_name='bc_func'),
            inputs=['bc_func_code', 'code_module'],
            outputs='bc_func_fn'),
], description='Generates and compiles the required BVP functions from problem data')



import numpy as np
from scipy.optimize import minimize
import scipy.integrate
import scipy.interpolate
import time
import copy
from beluga.utils import keyboard
from scipy.integrate import simps


class Algorithm(object):
    """
    Object representing an algorithm that solves initial valued problems.

    This object serves as a base class for other algorithms.
    """

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)

        return obj


class Propagator(Algorithm):
    """
    Propagator of differential equations.
    """

    def __new__(cls, *args, **kwargs):
        """
        Creates a new Propagator object.

        :param args: Unused
        :param kwargs: Additional parameters accepted by the solver.
        :return: Propagator object.

        +------------------------+-----------------+-----------------+
        | Valid kwargs           | Default Value   | Valid Values    |
        +========================+=================+=================+
        | abstol                 | 1e-6            |  > 0            |
        +------------------------+-----------------+-----------------+
        | maxstep                | 0.1             |  > 0            |
        +------------------------+-----------------+-----------------+
        | reltol                 | 1e-6            |  > 0            |
        +------------------------+-----------------+-----------------+
        """

        obj = super().__new__(cls, *args, **kwargs)
        obj.abstol = kwargs.get('abstol', 1e-6)
        obj.maxstep = kwargs.get('maxstep', 0.1)
        obj.reltol = kwargs.get('reltol', 1e-6)
        return obj

    def __call__(self, eom_func, quad_func, tspan, y0, q0, *args, **kwargs):
        """
        Propagates the differential equations over a defined time interval.

        :param eom_func: Function representing the equations of motion.
        :param quad_func: Function representing the quadratures.
        :param tspan: Independent time interval.
        :param y0: Initial state position.
        :param q0: Initial quad position.
        :param args: Additional arguments required by EOM files.
        :param kwargs: Unused.
        :return: A full reconstructed trajectory, :math:`\\gamma`.
        """

        int_sol = scipy.integrate.solve_ivp(lambda t, y: eom_func(t, y, *args), [tspan[0], tspan[-1]], y0,
                                            rtol=self.reltol, atol=self.abstol, max_step=self.maxstep)

        gamma = Trajectory(int_sol.t, int_sol.y.T)

        if quad_func is not None:
            gamma = reconstruct(quad_func, gamma, *args)

        return gamma


class Trajectory(object):
    """
    Class containing information for a trajectory.

    .. math::
        \\gamma(t) : I \\subset \\mathbb{R} \\rightarrow B
    """

    def __new__(cls, *args, **kwargs):
        """
        Creates a new Trajectory object.

        :param args: :math:`(t, y, q, u)`
        :param kwargs: Unused.
        :return: Trajectory object.
        """

        obj = super(Trajectory, cls).__new__(cls)
        obj.t = None
        obj.y = None
        obj.q = None
        obj.u = None

        interpolation_type = kwargs.get('interpolation_type', 'linear').lower()
        obj.interpolation_type = interpolation_type

        l = len(args)
        if l >= 1:
            obj.t = args[0]

        if l >= 2:
            obj.y = args[1]

        if l >= 3:
            obj.q = args[2]

        if l >= 4:
            obj.u = args[3]

        return obj

    def __init__(self, *args, **kwargs):
        self.interpolate = None
        self.set_interpolate_function(self.interpolation_type)

    def __call__(self, t):
        """
        Mapping function for a trajectory.

        :param t: Time as :math:`t \\in \\mathbb{R}`
        :return: Returns position values :math:`(y, q, u) \\in B`
        """

        t = np.array(t, dtype=np.float64)
        y_val = None
        q_val = None
        u_val = None

        if len(self.t) == 0:
            return None, None, None

        if len(self.y.shape) == 1:
            dim = 1
        else:
            dim = self.y.shape[1]

        # This builds the interpolation function on the most up to date data
        if dim == 1:
            f = self.interpolate(self.t, self.y.T)
            y_val = np.array([f(t)])
        else:
            f = [self.interpolate(self.t, self.y.T[ii]) for ii in range(dim)]
            y_val = np.array([f[ii](t) for ii in range(dim)]).T

        return y_val, q_val, u_val

    def set_interpolate_function(self, func):
        """
        Sets the function used for interpolation.

        :param func: Function for interpolation or a string for SciPy's interp1d(kind=func).
        """
        if callable(func):
            self.interpolate = func
        elif isinstance(func, str):
            func = func.lower()
            self.interpolate = lambda t, y: scipy.interpolate.interp1d(t, y, kind=func)

    def __getitem__(self, item):
        t_val = None
        y_val = None
        q_val = None
        u_val = None
        if self.t is not None:
            t_val = self.t[item]

        if self.y is not None:
            y_val = self.y[item]

        if self.q is not None:
            q_val = self.q[item]

        if self.u is not None:
            u_val = self.u[item]

        return t_val, y_val, q_val, u_val

    def __len__(self):
        if self.t is None:
            return 0
        else:
            return len(self.t)


def reconstruct(quadfun, gamma, *args):
    """
    Completely reconstructs a trajectory for all time in :math:`\\gamma`.

    .. math::
        \\begin{aligned}
            \\text{reconstruct} : \\gamma \\in B/Q &\\rightarrow \\gamma \\in B \\\\
            (g, \\gamma) &\\mapsto \\int_{t_0}^{t} g \\circ \\gamma dt \\; \\forall \\; t
        \\end{aligned}

    :param quadfun: Equations of motion on the symmetry space.
    :param gamma: Trajectory in quotient space :math:`B/Q`.
    :param args: Additional arguments needed by quadfun.
    :return: :math:`\\gamma` - Reconstructed trajectory in total space :math:`B`.
    """

    gamma = copy.copy(gamma)
    if gamma.q is None:
        q0 = 0
    else:
        q0 = gamma.q[0]

    l = len(gamma)
    temp_q = np.array([integrate_quads(quadfun, [gamma.t[0], gamma.t[0]], gamma, *args)])

    for ii in range(l-1):
        qf = integrate_quads(quadfun, [gamma.t[ii], gamma.t[ii+1]], gamma, *args)
        temp_q = np.vstack((temp_q, temp_q[-1] + qf))

    gamma.q = temp_q + q0
    return gamma


def integrate_quads(quadfun, tspan, gamma, *args):
    """
    Integrates quadratures over a trajectory base space. Only returns the terminal point.

    .. math::
        \\begin{aligned}
            \\text{integrate_quads} : \\gamma \\in B/Q &\\rightarrow q_f \\in B \\\\
            (g, \\gamma) &\\mapsto \\int_{t_0}^{t_f} g \\circ \\gamma dt
        \\end{aligned}

    :param quadfun: Equations of motion on the symmetry space.
    :param tspan: Time interval to integrate over.
    :param gamma: Trajectory in quotient space :math:`B/Q`.
    :param args: Additional arguments needed by quadfun.
    :return: Value of the quads at :math:`t_f`.
    """

    if tspan[0] < gamma.t[0]:
        raise Exception('Time span out of integration bounds.')

    if tspan[-1] > gamma.t[-1]:
        raise Exception('Time span out of integration bounds.')

    l = len(gamma.t)
    x_set_temp = np.arange(0,l,1)

    ind0 = int(np.ceil(np.interp(tspan[0], gamma.t, x_set_temp)))
    indf = int(np.ceil(np.interp(tspan[-1], gamma.t, x_set_temp)))

    if tspan[0] != gamma.t[ind0]:
        x_interp = np.array([tspan[0]])
    else:
        x_interp = np.array([])

    x_interp = np.hstack((x_interp, gamma.t[ind0:indf]))

    if tspan[-1] != gamma.t[indf-1]:
        x_interp = np.hstack((x_interp, tspan[-1]))

    y0, q0, u0 = gamma(x_interp[0])

    # Evaluate the quad function over every point in the given interval
    dq = np.array([quadfun(time, gamma(time)[0], *args) for time in x_interp])

    # Integrate the quad func using numerical quadrature
    qf_m0 = simps(dq.T, x=x_interp)

    # Add the initial state to get the final state.
    if q0 is None:
        q0 = 0

    qf = qf_m0 + q0

    return qf



#==============================================================================#
# Author: Justin Mansell (2016)
# Description: Continuation for 2 variables using a random path.
# Note: Many functions redundant with manual strategy. This is intentional
#       to avoid being dependent on ManualStrategy.py (which may be updated).
#==============================================================================#
from beluga.continuation.ContinuationVariable import ContinuationVariable
#from RandomPathGen import RandomPath
import numpy as np
import random
import logging

def RandomPath(Xi, Xf, minsteps, agg = 0.5):
      #Random Path generation routine.
      #INPUT: Xi = array of initial coordinates
      #       Xf = array of target coordinates
      #      agg = aggression parameter (0 < agg <= 1)
      # minsteps = minimum number of steps for the path
      #OUTPUT: X = array of step coordinates for each variable
      Xi = np.array(Xi)
      Xf = np.array(Xf)
      dl = np.linalg.norm(abs(Xf - Xi)/minsteps) #Compute nominal step length
      maxsteps = 1000 #Maximum number of steps
      alfa = (1 - agg) * np.pi/2.0 #Maximum angle from center line for step direction

      #-----------------------------Generate Path--------------------------------#
      step = 0 #Step counter
      X = [Xi] #List of step coordinates
      while step < maxsteps:
          Xc = X[step] #Current location
          delX = Xf - Xc #Target vector
          if np.linalg.norm(delX) <= dl:
              X.append(Xf) #Within a step length of target. We're done.
              break

          ##Create list of angles defining the target vector
          cline = [] #List of angles to define delX vector
          cline.append(np.arctan2(delX[1], delX[0]))

          ##Randomize each hyperspherical coordiante
          phi = []
          for theta in cline:
              phi.append(random.uniform(theta - alfa, theta + alfa))

          ##Compute the step
          dx = np.array([0.0] * len(Xi)) #Step length vector
          dx[0] = dl * np.cos(phi[0])
          dx[1] = dl * np.sin(phi[0])
          X.append(Xc + dx)
          step = step + 1
      return X



# Can be subclassed to allow automated stepping
class RandomStrategy(object):
    """Defines one continuation step in continuation set"""
    # A unique short name to select this class
    name = 'random'

    def __init__(self, num_cases = 1, vars=[], bvp=None):
        self.bvp = bvp
        self._num_cases = num_cases
        self._agg = 0.5 # aggression parameter (can range from 0 to 1)
        self.vars = {}  # dictionary of values
        self.ctr  = 0   # iteration counter
        self.last_bvp = None

    def reset(self):
        """Resets the internal step counter to zero"""
        self.ctr = 0
        self.last_bvp = None

    def clear(self):
        """Clears all the previously set continuation variables"""
        self.vars = {}
        self.reset()
    #
    # def get_ctr(self):
    #     return self.ctr

    def set_bvp(self, bvp):
        self.bvp = bvp
        #Check that we have the correct number of variables
        #TODO: Generalize random path to any number of continuation variables
        num_cont_vars = 0
        for var_type in self.vars.keys(): #Count all the continuation variables
            num_cont_vars+=len(self.vars[var_type])
        if num_cont_vars > 2:
            raise ValueError('Cannot yet run random continuation for more than two variables')

        # Iterate through all types of variables
        Xi=[] #Initial vector in the continuation space
        Xf=[] #Target vector in the continuation space
        for var_type in self.vars.keys():
            for var_name in self.vars[var_type].keys():

                # Look for the variable name from continuation in the BVP
                if var_name not in bvp.solution.aux[var_type].keys():
                    raise ValueError('Variable '+var_name+' not found in boundary value problem')

                # Set current value of each continuation variable
                self.vars[var_type][var_name].value = bvp.solution.aux[var_type][var_name]

                #Append the start and end states of the variable to X
                Xi.append(self.vars[var_type][var_name].value)
                Xf.append(self.vars[var_type][var_name].target)

        #Generate the continuation path
        Xpath = RandomPath(Xi,Xf,self._num_cases,self._agg)

        #Desseminate the steps for each variable
        varsteps1, varsteps2 = zip(*Xpath)
        for var_type in self.vars.keys():
            for var_name in self.vars[var_type].keys():
                if self.vars[var_type][var_name].target == varsteps1[-1]:
                    self.vars[var_type][var_name].steps=varsteps1
                else:
                    self.vars[var_type][var_name].steps=varsteps2

    def set(self, var_type,name,target):
        if var_type not in self.vars.keys():
            self.vars[var_type] = {}

        # Create continuation variable object
        self.vars[var_type][name] = ContinuationVariable(name,target)
        return self

    def num_cases(self,num_cases=None, agg=0.5):
        if num_cases is None:
            return self._num_cases
        else:
            if self.ctr > 0:
                raise RuntimeError('Cannot set num_cases during iteration')

            self._num_cases = num_cases
            self._agg   = agg
            return self

    def terminal(self, name,target):
        self.set('terminal',name,target)
        return self

    def initial(self, name,target):
        self.set('initial',name,target)
        return self

    def const(self, name,target):
        self.set('const',name,target)
        return self

    # def constraint(self, name,target):
    #     self.set('constraint',name,target)
    #     return self

    def __iter__(self):
        """Define class as being iterable"""
        return self

    def __next__(self):
        return self.next()

    def next(self, ignore_last_step = False):
        """Generator class to create BVPs for the continuation step iterations
        ignore_last_bvp: Should the non-convergence of previous step be ignored?
        """

        if self.bvp is None:
            raise ValueError('No boundary value problem associated with this object')

        if not ignore_last_step and self.last_bvp is not None and not self.last_bvp.solution.converged:
            logging.error('The last step did not converge!')
            raise RuntimeError('Solution diverged! Stopping.')

        #Check the actual number of cases (may be less than specified number
        #due to the random path)
        actual_num_cases=[]
        for var_type in self.vars.keys():
            for var_name in self.vars[var_type].keys():
                actual_num_cases.append(len(self.vars[var_type][var_name].steps))
        for ncases in actual_num_cases:
            if ncases != actual_num_cases[0]:
                raise ValueError('Step vector for different variables have different lengths')

        if self.ctr >= ncases:
            raise StopIteration

        # Update auxiliary variables using previously calculated step sizes
        for var_type in self.vars:
            for var_name in self.vars[var_type]:
                self.bvp.solution.aux[var_type][var_name] = self.vars[var_type][var_name].steps[self.ctr]

        self.ctr += 1
        self.last_bvp = self.bvp
        return self.bvp



from .BaseAlgorithm import BaseAlgorithm
from .Shooting import Shooting
from .Collocation import Collocation
from .QCPI import QCPI

import os
import glob
modules = glob.glob(os.path.dirname(__file__)+"/*.py")
__all__ = [os.path.basename(f)[:-3] for f in modules]



from beluga.visualization.renderers import BaseRenderer
from bokeh.plotting import *
from bokeh.palettes import *
from bokeh.models import HoverTool

import webbrowser

class Bokeh(BaseRenderer):
    def __init__(self, filename='plot.html'):
        self._figures = []
        self.filename = filename
        output_file(filename, title="beluga Output")


    def _get_figure(self,f):
        """
        Returns Figure instance from internal list using index

        Raises:
            ValueError if invalid index is used
        """
        try:
            fh =  self._figures[f]
            if fh is None:
                raise ValueError('Invalid figure handle specified!')
            return fh
        except:
            raise ValueError('Invalid figure handle specified!')

    def create_figure(self, width=600, height=600):
        """
        Creates a new figure and returns a handle (index into array)
        """
        self._figures.append(figure(width=width, height=height))
        return len(self._figures)-1

    def close_figure(self,f):
        """
        Closes a specified figure
        """
        pass
        # close(self._get_figure(f))

    def show_figure(self,f):
        """
        Shows a specified figure
        """
        show(self._get_figure(f))

    def show_all(self):
        """
        Show all rendered figures
        """
        p = vplot(*self._figures)
        show(p)
        # for f in self._figures:
        #
        #     show(f)
        webbrowser.open(self.filename, new=2, autoraise=True)

    def render_plot(self,f,p):
        """
        Adds a line plot using the given data to the specified figure
        """
        plot = self._get_figure(f);

        for line in p.plot_data:
            dataset = [(data['x_data'],data['y_data']) for data in line['data']]
            xlist, ylist = zip(*dataset)
            if len(dataset) < 5:
                mypalette=Spectral4[0:len(dataset)]
            else:
                mypalette=Spectral10[0:len(dataset)]
            plot.multi_line(xs=xlist, ys=ylist, line_color=mypalette)

        # NOT WORKING!
        # if p._xlabel is not None and p._ylabel is not None:
        #     hover = HoverTool(
        #         tooltips=[
        #             (p._xlabel+":", "$x"),
        #             (p._ylabel+":", "$y"),
        #         ]
        #     )
        # else:
        #     hover = HoverTool(
        #         tooltips=[
        #             ("($x, $y)")
        #         ]
        #     )
        # plot.add_tools(hover)

        if p._xlabel is not None:
            plot.xaxis.axis_label = p._xlabel
        if p._ylabel is not None:
            plot.yaxis.axis_label = p._ylabel
        if p._title is not None:
            plot.title = p._title

    def render_subplot(self,f,index,plot):
        """
        Adds a subplot to the specified figure
        """
        pass

if __name__ == '__main__':
    from beluga.visualization.elements import Plot
    import dill

    r = Bokeh()
    fig = r.create_figure()

    with open('/Users/tantony/dev/tantony-beluga/examples/planarHypersonic/phu_2k5_eps2.dill','rb') as f:
        out = dill.load(f)

    p = Plot(-1,-1)
    p.line('v/1000','h/1000')
    p.xlabel('v (km/s)')
    p.ylabel('h (km)')
    p.title('Altitude vs. Velocity')
    p.preprocess(out['solution'],out['problem_data'])

    r.render_plot(fig,p)
    r.show_figure(fig)

    fig = r.create_figure()
    p = Plot(-1,-1)
    p.line('theta*re/1000','h/1000')
    p.xlabel('downrange (km)')
    p.ylabel('altitude (km)')
    p.title('Altitude vs. Downrange')
    p.preprocess(out['solution'],out['problem_data'])
    r.render_plot(fig,p)
    r.show_figure(fig)



import abc
import dill
import logging

#TODO: Identify data source based on file extension
class BaseDataSource(object):
    __metaclass__ = abc.ABCMeta
    valid_exts = []
    @abc.abstractmethod
    def reset(self):
        """
        Resets data source
        """

    @abc.abstractmethod
    def load(self):
        """
        Loads data into memory
        """

    @abc.abstractmethod
    def get_problem(self):
        """
        Returns problem information
        """

    @abc.abstractmethod
    def get_solution(self):
        """
        Returns the solution array
        """


class Dill(BaseDataSource):
    valid_exts = ['dill']
    def __init__(self, filename = 'data.dill'):
        """
        Initializes the data source with supplied filename
        """
        self.is_loaded = False
        self.filename = filename

    def reset(self):
        self.is_loaded = False
        self._data = None

    def load(self):
        """
        Loads solution data using dill if not already loaded
        """
        if not self.is_loaded:
            with open(self.filename,'rb') as f:
                logging.info("Loading datafile "+self.filename+"...")
                self._data = dill.load(f)
                if 'solution' not in self._data:
                    self.is_loaded = False
                    logging.error("Solution missing in data file :"+self.filename)
                    raise RuntimeError("Solution missing in data file :"+self.filename)
                if 'problem_data' not in self._data:
                    self.is_loaded = False
                    logging.error("Problem data missing in data file :"+self.filename)
                    raise RuntimeError("Problem data missing in data file :"+self.filename)

                logging.info("Loaded "+str(len(self._data['solution']))+" solution sets from "+self.filename)
                self.is_loaded = True

    def get_problem(self):
        """
        Return problem data
        """
        # Lazy load data
        if not self.is_loaded:
            self.load()
            # logging.error("Data source should be loaded before being used")
            # raise RuntimeError("Data source should be loaded before being used")
        return self._data['problem_data']


    def get_solution(self):
        """
        Returns solution array
        """
        # Lazy load data
        if not self.is_loaded:
            self.load()
            # logging.error("Data source should be loaded before being used")
            # raise RuntimeError("Data source should be loaded before being used")
        return self._data['solution']

from scipy.io import loadmat
from beluga.bvpsol import Solution
import numpy as np
class GPOPS(BaseDataSource):
    valid_exts = ['mat']
    def __init__(self, filename = 'data.mat', states=None, controls=None, const=None):
        """
        Initializes the data source with supplied filename
        """
        if states is None or controls is None:
            raise ValueError('Please specify both state and control variable names')

        self.is_loaded = False
        self.filename = filename
        states = tuple(states)
        costates = tuple('lam'+x.upper() for x in states)
        self._const = const
        self.problem_data = {'state_list':states+costates,
                             'control_list':controls,
                             'quantity_vars':{}}
    def reset(self):
        self.is_loaded = False
        self._data = None

    def load(self):
        """
        Loads solution data using dill if not already loaded
        """
        if not self.is_loaded:
            logging.info("Loading datafile "+self.filename+"...")
            out = loadmat(self.filename)

            if 'output' in out:
                out = out['output']['result'][0][0][0][0]
            soldata = out['solution']['phase'][0][0][0][0]

            # if 'solution' not in self._data:
            #     self.is_loaded = False
            #     logging.error("Solution missing in data file :"+self.filename)
            #     raise RuntimeError("Solution missing in data file :"+self.filename)
            # if 'problem_data' not in self._data:
            #     self.is_loaded = False
            #     logging.error("Problem data missing in data file :"+self.filename)
            #     raise RuntimeError("Problem data missing in data file :"+self.filename)
            #
            _sol = Solution()

            tf = max(soldata['time'])
            _sol.x = soldata['time'][:,0]/tf
            _sol.y = np.r_[soldata['state'].T,soldata['costate'].T,np.ones_like(soldata['time']).T*tf]
            _sol.u = soldata['control'].T

            if 'tf' not in self.problem_data['state_list']:
                self.problem_data['state_list'] = tuple(self.problem_data['state_list']) + ('tf',)

            _sol.arcs = ((0, len(_sol.x)-1),)

            if self._const is not None:
                _sol.aux = {'const':self._const}

            self._sol = [[_sol]]
            logging.info('Loaded solution from data file')

            self.is_loaded = True

    def get_problem(self):
        """
        Return problem data
        """
        # Lazy load data
        if not self.is_loaded:
            self.load()
            # logging.error("Data source should be loaded before being used")
            # raise RuntimeError("Data source should be loaded before being used")
        return self.problem_data


    def get_solution(self):
        """
        Returns solution array
        """
        # Lazy load data
        if not self.is_loaded:
            self.load()
            # logging.error("Data source should be loaded before being used")
            # raise RuntimeError("Data source should be loaded before being used")
        return self._sol
